# Ingestion related functions which may be overriden by an implementing app,
# e.g. Foursight or CGAP portal, using the dcicutils project_utils mechanism.

class SnovaultProjectIngestion:
    def note_ingestion_enqueue_uuids_for_request(self, ingestion_type, request, uuids):
        pass



import pytest
from snovault.schema_utils import (
    _update_resolved_data, _handle_list_or_string_value,  # noQA - testing protected members
    resolve_merge_refs, validate
)


pytestmark = [pytest.mark.working, pytest.mark.schema]


def test_schema_utils_update_resolved_data(mocker):
    resolve_merge_ref = mocker.patch('snovault.schema_utils.resolve_merge_ref')

    def custom_resolver(ref, resolver):
        if ref == 'xyz':
            return {
                'a': 'new value',
                'and': 'a ref',
                'inside': 'of a ref',
                '$merge': 'notxyz',
            }
        else:
            return {
                'the': 'notxyz values',
                'with': {
                    'sub': 'dependencies',
                    'and': ['lists'],
                }
            }
    resolve_merge_ref.side_effect = custom_resolver
    resolved_data = {}
    _update_resolved_data(resolved_data, 'xyz', {})
    assert resolved_data == {
        'a': 'new value',
        'and': 'a ref',
        'inside': 'of a ref',
        'the': 'notxyz values',
        'with': {
            'sub': 'dependencies',
            'and': [
                'lists'
            ]
        }
    }


def test_schema_utils_handle_list_or_string_value(mocker):
    resolve_merge_ref = mocker.patch('snovault.schema_utils.resolve_merge_ref')

    def custom_resolver(ref, resolver):
        if ref == 'xyz':
            return {
                'a': 'new value',
                'and': 'a ref',
                'inside': 'of a ref',
                '$merge': 'notxyz',
            }
        else:
            return {
                'the': 'notxyz values',
                'with': {
                    'sub': 'dependencies',
                    'and': ['lists'],
                }
            }
    resolve_merge_ref.side_effect = custom_resolver
    resolved_data = {}
    value = 'notxyz'
    _handle_list_or_string_value(resolved_data, value, {})
    assert resolved_data == {
        'the': 'notxyz values',
        'with': {
            'sub': 'dependencies',
            'and': [
                'lists'
            ]
        }
    }
    resolved_data = {}
    value = ['notxyz', 'xyz']
    _handle_list_or_string_value(resolved_data, value, {})
    assert resolved_data == {
        'the': 'notxyz values',
        'with': {
            'sub': 'dependencies',
            'and': ['lists']
        },
        'a': 'new value',
        'and': 'a ref',
        'inside': 'of a ref'
    }

    def custom_resolver(ref, resolver):
        if ref == 'xyz':
            return {
                'a': 'b',
            }
        else:
            return {
                'c': 'd'
            }
    resolve_merge_ref.side_effect = custom_resolver
    resolved_data = {}
    value = ['notxyz', 'xyz']
    _handle_list_or_string_value(resolved_data, value, {})
    assert resolved_data == {
        'a': 'b',
        'c': 'd',
    }

    def custom_resolver(ref, resolver):
        if ref == 'xyz':
            return {
                'a': 'b',
            }
        else:
            return {
                'a': 'override'
            }
    resolve_merge_ref.side_effect = custom_resolver
    resolved_data = {}
    value = ['notxyz', 'xyz']
    _handle_list_or_string_value(resolved_data, value, {})
    assert resolved_data == {
        'a': 'b',
    }
    resolved_data = {}
    value = ['xyz', 'notxyz']
    _handle_list_or_string_value(resolved_data, value, {})
    assert resolved_data == {
        'a': 'override',
    }


def test_schema_utils_resolve_merge_refs_returns_copy_of_original_if_no_refs(mocker):
    resolver = None
    data = {'a': 'b'}
    resolved_data = resolve_merge_refs(data, resolver)
    assert resolved_data == data
    assert id(resolved_data) != id(data)
    data = {
        'a': 'b',
        'c': ['d', 'e', 1],
        'x': {
            'y': {
                'z': [
                    {
                        'p': 'r'
                    },
                    3.2,
                    True,
                    False,
                    None
                ]
            }
        }
    }
    resolved_data = resolve_merge_refs(data, resolver)
    assert resolved_data == data
    # Dicts are copies.
    assert id(resolved_data) != id(data)
    # Lists are copies.
    assert id(resolved_data['x']['y']['z'][0]) != id(data['x']['y']['z'][0])
    # Assignment doesn't reflect in original.
    resolved_data['x']['y']['z'][0]['p'] = 'new value'
    assert data['x']['y']['z'][0]['p'] == 'r'
    data = [
        'k',
        '5',
        '6',
        2,
        {},
    ]
    resolved_data = resolve_merge_refs(data, resolver)
    assert resolved_data == data
    assert id(resolved_data) != id(data)


def test_schema_utils_resolve_merge_refs_fills_in_refs(mocker):
    resolver = None
    data = {'a': 'b'}
    resolved_data = resolve_merge_refs(data, resolver)
    assert resolved_data == data
    resolve_merge_ref = mocker.patch('snovault.schema_utils.resolve_merge_ref')
    resolve_merge_ref.return_value = {'a new value': 'that was resolved'}
    data = {'a': 'b', 'c': {'$merge': 'xyz'}}
    resolved_data = resolve_merge_refs(data, resolver)
    expected_data = {'a': 'b', 'c': {'a new value': 'that was resolved'}}
    assert resolved_data == expected_data
    data = {
        'a': 'b',
        'c': {'$merge': 'xyz'},
        'sub': {
            'values': [
                'that',
                'were',
                'resolved',
                {
                    'if': {
                        '$merge': 'xyz',
                        'and': 'other',
                        'values': 'are',
                        'allowed': 'too',
                    }
                }
            ]
        }
    }
    resolved_data = resolve_merge_refs(data, resolver)
    expected_data = {
        'a': 'b',
        'c': {'a new value': 'that was resolved'},
        'sub': {
            'values': [
                'that',
                'were',
                'resolved',
                {
                    'if': {
                        'a new value': 'that was resolved',
                        'and': 'other',
                        'values': 'are',
                        'allowed': 'too',
                    }
                }
            ]
        }
    }

    def custom_resolver(ref, resolver):
        if ref == 'xyz':
            return {
                'a': 'new value',
                'and': 'a ref',
                'inside': 'of a ref',
                '$merge': 'notxyz',
            }
        else:
            return {
                'the': 'notxyz values',
                'with': {
                    'sub': 'dependencies',
                    'and': ['lists'],
                }
            }
    resolve_merge_ref.side_effect = custom_resolver
    data = {
        'something_new': [
            {
                '$merge': 'notxyz'
            }
        ],
        'a': 'b',
        'c': {'$merge': 'xyz'},
        'sub': {
            'values': [
                'that',
                'were',
                'resolved',
                {
                    'if': {
                        '$merge': 'xyz',
                        'and': 'other',
                        'values': 'are',
                        'allowed': 'too',
                    }
                }
            ]
        }
    }
    resolved_data = resolve_merge_refs(data, resolver)
    expected_data = {
        'something_new': [
            {
                'the': 'notxyz values',
                'with': {
                    'sub': 'dependencies',
                    'and': ['lists']
                }
            }
        ],
        'a': 'b',
        'c': {
            'a': 'new value',
            'and': 'a ref',
            'inside': 'of a ref',
            'the': 'notxyz values',
            'with': {
                'sub': 'dependencies',
                'and': ['lists']
            }
        },
        'sub': {
            'values': [
                'that', 'were',
                'resolved', {
                    'if': {
                        'a': 'new value',
                        'and': 'other',
                        'inside': 'of a ref',
                        'the': 'notxyz values',
                        'with': {
                            'sub': 'dependencies',
                            'and': ['lists']
                        },
                        'values': 'are',
                        'allowed': 'too'
                    }
                }
            ]
        }
    }
    assert resolved_data == expected_data


def test_schema_utils_resolve_merge_refs_fills_allows_override_of_ref_property(mocker):
    resolver = None
    resolve_merge_ref = mocker.patch('snovault.schema_utils.resolve_merge_ref')
    resolve_merge_ref.return_value = {
        'a new value': 'that was resolved',
        'custom': 'original value',
        'and': 'something else',
    }
    data = {
        'a': {
            '$merge': 'xyz',
            'custom': 'override',
        }
    }
    resolved_data = resolve_merge_refs(data, resolver)
    expected_data = {
        'a': {
            'a new value': 'that was resolved',
            'and': 'something else',
            'custom': 'override',
        }
    }
    assert resolved_data == expected_data


def test_schema_utils_resolve_merge_ref_in_real_schema(testapp):
    """ Tests that we can resolve a $merge object in our test schema """
    collection_url = '/testing-linked-schema-fields'
    # test posting $merge quality field
    testapp.post_json(collection_url, {'quality': 5}, status=201)
    testapp.post_json(collection_url, {'quality': 'no'}, status=422)
    # test posting $merge linkTo
    atid = testapp.post_json('/testing-link-targets-sno', {'name': 'target'}, status=201).json['@graph'][0]['@id']
    testapp.post_json(collection_url, {'quality': 5,
                                       'linked_targets': [
                                           {'target': atid, 'test_description': 'test'}
                                       ]}, status=201)

def test_schema_utils_resolve_merge_ref_in_embedded_schema(testapp):
    """ Tests that we can resolve a $merge object in our test schema embedded in another schema """
    collection_url = '/testing-linked-schema-fields'
    atid = testapp.post_json('/testing-link-targets-sno', {'name': 'target'}, status=201).json['@graph'][0]['@id']
    atid = testapp.post_json(collection_url, {'quality': 5,
                                              'linked_targets': [
                                                  {'target': atid, 'test_description': 'test'}
                                              ]}, status=201).json['@graph'][0]['@id']
    atid = testapp.post_json('/testing-embedded-linked-schema-fields', {
        'link': atid
    }, status=201).json['@graph'][0]['@id']
    embedded = testapp.get(f'/{atid}?frame=embedded', status=200).json
    assert embedded['link']['quality'] == 5
    assert embedded['link']['linked_targets'][0]['test_description'] == 'test'


@pytest.mark.parametrize('invalid_date', [
    'not a date',
    'also-not-a-date',
    '10-5-2022',
    '10-05-almost',
    '10-05-2023f'
])
def test_schema_utils_validates_dates(testapp, invalid_date):
    """ Tests that our validator will validate dates """
    schema = {
        "type": "object",
        "properties": {
            "date_property": {
                "type": "string",
                "format": "date"
            }
        }
    }
    _, errors = validate(schema, {
        'date_property': invalid_date
    })
    date_error = str(errors[0])
    assert f"{invalid_date!r} is not a 'date'" in date_error


@pytest.mark.parametrize('invalid_date_time', [
    'not a date',
    'also-not-a-date',
    '10-5-2022',
    '10-05-almost',
    '10-05-2023f',
    '1424-45-93T15:32:12.9023368Z',
    '20015-10-23T15:32:12.9023368Z',
    '2001-130-23T15:32:12.9023368Z',
    '2001-10-233T15:32:12.9023368Z',
    '2001-10-23T153:32:12.9023368Z'
])
def test_schema_utils_validates_date_times(testapp, invalid_date_time):
    """ Tests that our validator will validate date-time """
    schema = {
        "type": "object",
        "properties": {
            "date_time_property": {
                "type": "string",
                "format": "date-time"
            }
        }
    }
    _, errors = validate(schema, {
        'date_time_property': invalid_date_time
    })
    date_error = str(errors[0])
    assert f"{invalid_date_time!r} is not a 'date-time'" in date_error



import base64
import datetime
import jwt
import os
import requests
import structlog

from dateutil.parser import isoparse
from dcicutils.misc_utils import remove_element, ignorable, ignored
from operator import itemgetter
from passlib.context import CryptContext
from pyramid.authentication import (
    BasicAuthAuthenticationPolicy as _BasicAuthAuthenticationPolicy,
    CallbackAuthenticationPolicy
)
from pyramid.httpexceptions import HTTPForbidden, HTTPUnauthorized
from pyramid.path import DottedNameResolver, caller_package
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.view import view_config
from snovault import ROOT, COLLECTIONS
from snovault.calculated import calculate_properties
from snovault.crud_views import collection_add as sno_collection_add
from snovault.project_app import app_project
from snovault.schema_utils import validate_request
from snovault.util import debug_log
from snovault.validation import ValidationFailure
from snovault.validators import no_validate_item_content_post
from urllib.parse import urlencode


log = structlog.getLogger(__name__)


CRYPT_CONTEXT = __name__ + ':crypt_context'


JWT_ENCODING_ALGORITHM = 'HS256'

# Might need to keep a list of previously used algorithms here, not just the one we use now.
# Decryption algorithm used to default to a long list, but more recent versions of jwt library
# say we should stop assuming that.
#
# In case it goes away, as far as I can tell, the default for decoding from their
# default_algorithms() method used to be what we've got in JWT_ALL_ALGORITHMS here.
#  -kmp 15-May-2020

JWT_ALL_ALGORITHMS = ['ES512', 'RS384', 'HS512', 'ES256', 'none',
                      'RS256', 'PS512', 'ES384', 'HS384', 'ES521',
                      'PS384', 'HS256', 'PS256', 'RS512']

# Probably we could get away with fewer, but I think not as few as just our own encoding algorithm,
# so for now I believe the above list was the default, and this just rearranges it to prefer the one
# we use for encoding. -kmp 19-Jan-2021

JWT_DECODING_ALGORITHMS = [JWT_ENCODING_ALGORITHM] + remove_element(JWT_ENCODING_ALGORITHM, JWT_ALL_ALGORITHMS)


def includeme(config):
    config.include('.edw_hash')
    setting_prefix = 'passlib.'
    passlib_settings = {
        k[len(setting_prefix):]: v
        for k, v in config.registry.settings.items()
        if k.startswith(setting_prefix)
    }
    if not passlib_settings:
        passlib_settings = {'schemes': 'edw_hash, unix_disabled'}
    crypt_context = CryptContext(**passlib_settings)
    config.registry[CRYPT_CONTEXT] = crypt_context

    # basic login route
    config.add_route('login', '/login')
    config.add_route('logout', '/logout')
    config.add_route('me', '/me')
    config.add_route('impersonate-user', '/impersonate-user')
    config.add_route('session-properties', '/session-properties')
    config.add_route('create-unauthorized-user', '/create-unauthorized-user')
    config.scan(__name__)


class NamespacedAuthenticationPolicy(object):
    """ Wrapper for authentication policy classes

    As userids are included in the list of principals, it seems good practice
    to namespace them to avoid clashes.

    Constructor Arguments

    ``namespace``

        The namespace used (string).

    ``base``

        The base authentication policy (class or dotted name).

    Remaining arguments are passed to the ``base`` constructor.

    Example

    To make a ``REMOTE_USER`` 'admin' be 'user.admin'

    .. code-block:: python

        policy = NamespacedAuthenticationPolicy('user',
            'pyramid.authentication.RemoteUserAuthenticationPolicy')
    """

    def __new__(cls, namespace, base, *args, **kw):
        # Dotted name support makes it easy to configure with pyramid_multiauth
        name_resolver = DottedNameResolver(caller_package())
        base = name_resolver.maybe_resolve(base)
        # Dynamically create a subclass
        name = 'Namespaced_%s_%s' % (namespace, base.__name__)
        klass = type(name, (cls, base), {'_namespace_prefix': namespace + '.'})
        return super(NamespacedAuthenticationPolicy, klass).__new__(klass)

    def __init__(self, namespace, base, *args, **kw):
        ignored(namespace, base)  # TODO: SHOULD this be ignored?
        super().__init__(*args, **kw)

    def unauthenticated_userid(self, request):
        return app_project().namespaced_authentication_policy_unauthenticated_userid(self, request)

    def _unauthenticated_userid_implementation(self, request):
        userid = super().unauthenticated_userid(request)
        if userid is not None:
            userid = self._namespace_prefix + userid
        return userid

    def authenticated_userid(self, request, set_user_info_property=True):
        # TODO: Maybe something like ...
        # return app_project().login_policy.authenticated_userid(request, set_user_info_property)
        return app_project().namespaced_authentication_policy_authenticated_userid(self, request, set_user_info_property)

    def _authenticated_userid_implementation(self, request, set_user_info_property=True):
        """
        Adds `request.user_info` for all authentication types.
        Fetches and returns some user details if called.
        """
        namespaced_userid = super().authenticated_userid(request)

        if not set_user_info_property:
            return namespaced_userid

        if namespaced_userid is not None:
            # userid, if present, may be in form of UUID (if remoteuser) or an email (if Auth0).
            namespace, userid = namespaced_userid.split(".", 1)

            # Allow access basic user credentials from request obj after authenticating & saving request
            def get_user_info(request):
                user_props = request.embed('/session-properties', as_user=userid)  # Performs an authentication against DB for user.
                if not user_props.get('details'):
                    raise HTTPUnauthorized(
                        title="Could not find user info for {}".format(userid),
                        headers={
                            'WWW-Authenticate':
                                "Bearer realm=\"{}\"; Basic realm=\"{}\"".format(request.domain, request.domain)
                        }
                    )
                return user_props

            # If not authenticated (not in our DB), request.user_info will throw an HTTPUnauthorized error.
            request.set_property(get_user_info, "user_info", True)

        return namespaced_userid

    def remember(self, request, principal, **kw):
        if not principal.startswith(self._namespace_prefix):
            return []
        principal = principal[len(self._namespace_prefix):]
        return super().remember(request, principal, **kw)


class BasicAuthAuthenticationPolicy(_BasicAuthAuthenticationPolicy):
    def __init__(self, check, *args, **kw):
        # Dotted name support makes it easy to configure with pyramid_multiauth
        name_resolver = DottedNameResolver(caller_package())
        check = name_resolver.maybe_resolve(check)
        super().__init__(check, *args, **kw)


class LoginDenied(HTTPUnauthorized):
    title = 'Login Failure'

    def __init__(self, domain=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.headers.get('WWW-Authenticate') and domain:
            # headers['WWW-Authenticate'] might be set in constructor thru headers
            self.headers['WWW-Authenticate'] = "Bearer realm=\"{}\"; Basic realm=\"{}\"".format(domain, domain)


_fake_user = object()


class Auth0AuthenticationPolicy(CallbackAuthenticationPolicy):

    login_path = '/login'
    method = 'POST'

    def unauthenticated_userid(self, request):
        """
        So basically this is used to do a login, instead of the actual
        login view... not sure why, but yeah..
        """

        # we will cache it for the life of this request, cause pyramids does traversal
        cached = getattr(request, '_auth0_authenticated', _fake_user)

        if cached is not _fake_user:
            return cached

        # try to find the token in the request (should be in the header)
        id_token = get_jwt(request)
        if not id_token:
            # can I thrown an 403 here?
            # print('Missing assertion.', 'unauthenticated_userid', request)
            return None

        jwt_info = self.get_token_info(id_token, request)
        if not jwt_info:
            return None

        email = request._auth0_authenticated = jwt_info['email'].lower()

        # At this point, email has been authenticated with their Auth0 provider and via `get_token_info`,
        # but we don't know yet if this email is in our database. `authenticated_userid` should take care of this.

        app_project().note_auth0_authentication_policy_unauthenticated_userid(self, request, email, id_token)

        return email

    @staticmethod
    def email_is_partners_or_hms(payload):
        """
        Checks that the given JWT payload belongs to a partners email.
        """
        for identity in payload.get('identities', []):  # if auth0 decoded
            if identity.get('connection', '') in ['partners', 'hms-it']:
                return True

        # XXX: Refactor to use regex? Also should potentially be data-driven?
        if 'partners' in payload.get('sub', ''):
            return True
        elif 'harvard.edu' in payload.get('sub', ''):
            return True
        elif payload.get('email_verified'):
            return True
        else:
            return False

    @staticmethod
    def get_token_info(token, request):
        """
        Given a jwt get token info from auth0, handle retrying and whatnot.
        This is only called if we receive a Bearer token in Authorization header.
        """
        try:
            # lets see if we have an auth0 token or our own
            registry = request.registry
            auth0_client = registry.settings.get('auth0.client')
            auth0_secret = registry.settings.get('auth0.secret')
            if auth0_client and auth0_secret:
                # leeway accounts for clock drift between us and auth0
                payload = jwt.decode(token, auth0_secret,
                                     algorithms=JWT_DECODING_ALGORITHMS,
                                     audience=auth0_client, leeway=30)
                if 'email' in payload and Auth0AuthenticationPolicy.email_is_partners_or_hms(payload):
                    request.set_property(lambda r: False, 'auth0_expired')
                    return payload

            else:  # we don't have the key, let auth0 do the work for us
                warn_msg = "No Auth0 keys present - falling back to making outbound network request to have Auth0 validate for us"
                log.warning(warn_msg)
                user_url = "https://{domain}/tokeninfo".format(domain='hms-dbmi.auth0.com')
                resp = requests.post(user_url, {'id_token': token})
                payload = resp.json()
                if 'email' in payload and Auth0AuthenticationPolicy.email_is_partners_or_hms(payload):
                    request.set_property(lambda r: False, 'auth0_expired')
                    return payload

        except jwt.exceptions.ExpiredSignatureError as e:
            ignorable(e)
            # Normal/expected expiration.

            # Allow us to return 403 code &or unset cookie in renderers.py
            request.set_property(lambda r: True, 'auth0_expired')

            return None

        except (ValueError, jwt.exceptions.InvalidTokenError, jwt.exceptions.InvalidKeyError) as e:
            # Catch errors from decoding JWT or unauthorized users.
            print('Invalid JWT assertion : %s (%s)' % (e, type(e).__name__))
            log.error("Error with JWT token (now unset) - " + str(e))
            request.set_property(lambda r: True, 'auth0_expired')  # Allow us to return 403 code &or unset cookie in renderers.py
            return None

        print("didn't get email or email is not verified")
        return None


def get_jwt_from_auth_header(request):
    if "Authorization" in request.headers:
        try:
            # Ensure this is a JWT token, not basic auth.
            # Per https://developer.mozilla.org/en-US/docs/Web/HTTP/Authentication and
            # https://tools.ietf.org/html/rfc6750, JWT is introduced by 'bearer', as in
            #   Authorization: Bearer something.something.something
            # rather than, for example, the 'basic' key information, which as discussed in
            # https://tools.ietf.org/html/rfc7617 is base64 encoded and looks like:
            #   Authorization: Basic QWxhZGRpbjpvcGVuIHNlc2FtZQ==
            # See also https://jwt.io/introduction/ for other info specific to JWT.
            [auth_type, auth_data] = request.headers['Authorization'].strip().split(' ', 1)
            if auth_type.lower() == 'bearer':
                return auth_data.strip()  # The spec says exactly one space, but then a token, so spaces don't matter
        except Exception:
            return None
    return None


def get_jwt(request):

    # First try to obtain JWT from headers (case: some REST API requests)
    token = get_jwt_from_auth_header(request)

    # If the JWT is not in the headers, get it from cookies (case: AJAX requests from portal & other clients)
    if not token:
        token = request.cookies.get('jwtToken')

    return token


@view_config(route_name='login', request_method='POST', permission=NO_PERMISSION_REQUIRED)
@debug_log
def login_view(context, request, samesite: str = "strict"):
    return app_project().login(context, request, samesite=samesite)


def login(context, request, *, samesite: str = "strict"):
    """
    Save JWT as httpOnly cookie
    """
    ignored(context)

    # Allow providing token thru Authorization header as well as POST request body.
    # Should be about equally secure if using HTTPS.
    request_token = get_jwt_from_auth_header(request)
    if request_token is None:
        request_token = request.json_body.get("id_token", None)

    is_https = (request.scheme == "https")

    request.response.set_cookie(
        "jwtToken",
        value=request_token,
        domain=request.domain,
        path="/",
        httponly=True,
        samesite=samesite,
        overwrite=True,
        secure=is_https
    )

    return {"saved_cookie": True}


@view_config(route_name='logout',
             permission=NO_PERMISSION_REQUIRED, http_cache=0)
@debug_log
def logout_view(context, request):
    return app_project().logout(context, request)


def logout(context, request):
    """
    This endpoint proxies a request to Auth0 for it to remove its session cookies.
    See https://auth0.com/docs/api/authentication#enterprise-saml-and-others-

    The Auth0 endpoint is meant to be navigated to by end-user as part of SSO logout (?)
    So this endpoint may not be needed at moment. Kept for reference.

    The front-end handles logging out by discarding the locally-held JWT from
    browser cookies and re-requesting the current 4DN URL.
    """
    ignored(context)

    # Deletes the cookie
    request.response.set_cookie(
        name='jwtToken',
        value=None,
        domain=request.domain,
        max_age=0,
        path='/',
        overwrite=True
    )

    request.response.status_code = 401
    request.response.headers['WWW-Authenticate'] = (
        "Bearer realm=\"{}\", title=\"Session Expired\"; Basic realm=\"{}\""
        .format(request.domain, request.domain)
    )

    return {"deleted_cookie": True}

    # TODO: NEED DO THIS CLIENTSIDE SO IT UNSETS USER'S COOKIE - MUST BE THRU REDIRECT NOT AJAX
    # (we don't do this - i.e. we don't bother to log user out of all of Auth0 session, just out of
    # own web app)

    # call auth0 to logout -
    # auth0_logout_url = "https://{domain}/v2/logout" \
    #             .format(domain='hms-dbmi.auth0.com')

    # requests.get(auth0_logout_url)

    # if asbool(request.params.get('redirect', True)):
    #     raise HTTPFound(location=request.resource_path(request.root))

    # return {}


@view_config(route_name='me', request_method='GET', permission=NO_PERMISSION_REQUIRED)
@debug_log
def me(context, request):
    """Alias /users/<uuid-of-current-user>"""
    ignored(context)
    for principal in request.effective_principals:
        if principal.startswith('userid.'):
            break
    else:
        raise HTTPForbidden(title="Not logged in.")

    namespace, userid = principal.split('.', 1)

    # return { "uuid" : userid } # Uncomment and delete below code to just grab UUID.

    request.response.status_code = 307  # Prevent from creating 301 redirects that get cached permanently by browser
    properties = request.embed('/users/' + userid, as_user=userid)
    return properties


def get_basic_properties_for_user(request, userid):
    user = request.registry[COLLECTIONS]['user'][userid]
    user_dict = user.__json__(request)

    # Only include certain/applicable fields from profile
    include_detail_fields = ['email', 'first_name', 'last_name', 'groups', 'timezone', 'status', 'project_roles']
    user_actions = calculate_properties(user, request, category='user_action')

    properties = {
        # 'user': request.embed(request.resource_path(user)),
        'details': {p: v for p, v in user_dict.items() if p in include_detail_fields},
        'user_actions': [v for k, v in sorted(user_actions.items(), key=itemgetter(0))]
    }

    # add uuid to user details
    properties['details']['uuid'] = userid

    return properties


@view_config(route_name='session-properties', request_method='GET',
             permission=NO_PERMISSION_REQUIRED)
@debug_log
def session_properties(context, request):
    ignored(context)
    for principal in request.effective_principals:
        if principal.startswith('userid.'):
            break
    else:
        # NOTE: returning details below allows internal remoteuser (TEST for example) to run DELETE requests
        # previously in downstream portal applications, the LoginDenied error was raised, preventing such
        # DELETE requests from occurring within unit testing. This can be re-enabled if desired in downstream
        # applications, but for now should stay like this so we can unit test DELETEs - Will April 6 2023
        if 'group.admin' in request.effective_principals:
            return {
                'details': {
                    'groups': [
                        'admin'
                    ]
                }
            }
        else:
            raise LoginDenied(domain=request.domain)

    namespace, userid = principal.split('.', 1)
    properties = get_basic_properties_for_user(request, userid)

    # if 'auth.userid' in request.session:
    #     properties['auth.userid'] = request.session['auth.userid']

    return properties


def basic_auth_check(username, password, request):
    """ This function implements the functionality that does the actual checking of the
        access key against what is in the database. It is thus very important. Access
        key expiration is implemented here - auth will fail if it has expired
    """
    # We may get called before the context is found and the root set
    root = request.registry[ROOT]
    collection = root['access-keys']
    try:
        access_key = collection[username]
    except KeyError:
        return None

    # Check expiration first
    # Note that access keys generated awhile ago will remain valid (for now) - will 6/14/21
    properties = access_key.properties
    expiration_date = properties.get('expiration_date')
    if expiration_date:
        dt = isoparse(expiration_date)  # datetime.date.fromisoformat in Python3.7
        now = datetime.datetime.utcnow()
        if now > dt:
            return None

    # If expiration valid, check hash
    hash = properties['secret_access_key_hash']
    crypt_context = request.registry[CRYPT_CONTEXT]
    valid = crypt_context.verify(password, hash)
    if not valid:
        return None

    return []  # success


@view_config(route_name='impersonate-user', request_method='POST',
             validators=[no_validate_item_content_post],
             permission='impersonate')
@debug_log
def impersonate_user(context, request):
    """As an admin, impersonate a different user."""
    ignored(context)

    userid = request.validated['userid']
    users = request.registry[COLLECTIONS]['user']

    try:
        user = users[userid]
    except KeyError:
        raise ValidationFailure('body', ['userid'], 'User not found.')

    if user.properties.get('status') != 'current':
        raise ValidationFailure('body', ['userid'], 'User is not enabled.')

    user_properties = get_basic_properties_for_user(request, userid)
    # pop off impersonate user action if not admin
    user_properties['user_actions'] = [x for x in user_properties['user_actions'] if (x['id'] and x['id'] != 'impersonate')]
    # make a key
    registry = request.registry
    auth0_client = registry.settings.get('auth0.client')
    auth0_secret = registry.settings.get('auth0.secret')
    if not (auth0_client and auth0_secret):
        raise HTTPForbidden(title="No keys to impersonate user")

    jwt_contents = {
        'email': userid,
        'email_verified': True,
        'aud': auth0_client,
    }

    id_token = jwt.encode(
        jwt_contents,
        auth0_secret,
        algorithm=JWT_ENCODING_ALGORITHM
    )

    is_https = request.scheme == "https"
    token_value = id_token.decode('utf-8') if isinstance(id_token, bytes) else id_token
    request.response.set_cookie(
        "jwtToken",
        value=token_value,
        domain=request.domain,
        path="/",
        httponly=True,
        samesite="strict",
        overwrite=True,
        secure=is_https
    )

    return user_properties


def generate_user():
    """ Generate a random user name with 64 bits of entropy
        Used to generate access_key
    """
    # Take a random 5 char binary string (80 bits of
    # entropy) and encode it as upper cased base32 (8 chars)
    random_bytes = os.urandom(5)
    user = base64.b32encode(random_bytes).decode('ascii').rstrip('=').upper()
    return user


def generate_password():
    """ Generate a password with 80 bits of entropy
    """
    # Take a random 10 char binary string (80 bits of
    # entropy) and encode it as lower cased base32 (16 chars)
    random_bytes = os.urandom(10)
    password = base64.b32encode(random_bytes).decode('ascii').rstrip('=').lower()
    return password


@view_config(route_name='create-unauthorized-user', request_method='POST',
             permission=NO_PERMISSION_REQUIRED)
@debug_log
def create_unauthorized_user(context, request):
    """
    Endpoint that creates an unauthorized user - so we can distinguish between those added by admins
    and through this API.
    For CGAP, an "unauthorized user" has cgap-core project association and nothing else.
    Requires a reCAPTCHA response, which is propogated from the front end
    registration form. This is so the endpoint cannot be abused.
    TODO: propagate key, secret from GAC

    Given a user properties in the request body, will validate those and also
    validate the reCAPTCHA response using the reCAPTCHA server. If all checks
    are successful, POST a new user and login

    Args:
        context: (ignored)
        request: Request object

    Returns:
        dictionary User creation response from collection_add

    Raises:
        LoginDenied, HTTPForbidden, or ValidationFailure
    """
    ignored(context)
    # env check
    env_name = request.registry.settings.get('env.name')
    if not app_project().env_allows_auto_registration(env_name):
        raise LoginDenied(f'Tried to register on {env_name} but it is disallowed')

    recaptcha_resp = request.json.get('g-recaptcha-response')
    if not recaptcha_resp:
        raise LoginDenied(f'Did not receive response from recaptcha!')

    email = request._auth0_authenticated  # equal to: jwt_info['email'].lower()
    user_props = request.json
    user_props_email = user_props.get("email", "<no e-mail supplied>").lower()
    if user_props_email != email:
        raise HTTPUnauthorized(
            title="Provided email {} not validated with Auth0. Try logging in again.".format(user_props_email),
            headers={
                'WWW-Authenticate':
                    "Bearer realm=\"{}\"; Basic realm=\"{}\"".format(request.domain, request.domain)}
        )

    # set user insert props
    del user_props['g-recaptcha-response']
    user_props['was_unauthorized'] = True
    user_props['email'] = user_props_email  # lower-cased
    user_coll = request.registry[COLLECTIONS]['User']
    request.remote_user = 'EMBED'  # permission = restricted_fields

    # validate the User json
    validate_request(user_coll.type_info.schema, request, user_props)
    if request.errors:
        raise ValidationFailure('body', 'create_unauthorized_user', 'Cannot validate request')

    # validate recaptcha_resp
    # https://developers.google.com/recaptcha/docs/verify
    recap_url = 'https://www.google.com/recaptcha/api/siteverify'
    recap_values = {
        'secret': request.registry.settings['g.recaptcha.secret'],
        'response': recaptcha_resp
    }
    data = urlencode(recap_values).encode()
    headers = {"Content-Type": "application/x-www-form-urlencoded; charset=utf-8"}
    recap_res = requests.get(recap_url, params=data, headers=headers).json()

    if recap_res['success']:
        sno_res = sno_collection_add(user_coll, request, False)  # POST User
        if sno_res.get('status') == 'success':
            return sno_res
        else:
            raise HTTPForbidden(title="Could not create user. Try logging in again.")
    else:
        # error with re-captcha
        raise HTTPUnauthorized(
            title="Invalid reCAPTCHA. Try logging in again.",
            headers={
                'WWW-Authenticate':
                    "Bearer realm=\"{}\"; Basic realm=\"{}\"".format(request.domain, request.domain)}
        )



import copy
import pytest
import random
import re

from dcicutils.misc_utils import override_environ
from ..util import (
    dictionary_lookup, DictionaryKeyError, merge_calculated_into_properties, CachedField,
    generate_indexer_namespace_for_testing,
)


def test_generate_indexer_namespace_for_testing():

    with override_environ(TRAVIS_JOB_ID="17", TEST_JOB_ID=None):
        ns = generate_indexer_namespace_for_testing()
        assert ns == "sno-test-17-"

        ns = generate_indexer_namespace_for_testing('foo')
        assert ns == "foo-test-17-"

    with override_environ(TRAVIS_JOB_ID=None, TEST_JOB_ID="17"):
        ns = generate_indexer_namespace_for_testing()
        assert ns == "sno-test-17-"

        ns = generate_indexer_namespace_for_testing('foo')
        assert ns == "foo-test-17-"

    with override_environ(TRAVIS_JOB_ID="17", TEST_JOB_ID="18"):
        ns = generate_indexer_namespace_for_testing()
        assert ns == "sno-test-18-"

        ns = generate_indexer_namespace_for_testing('foo')
        assert ns == "foo-test-18-"

    with override_environ(TEST_JOB_ID="snow-test-19"):
        ns = generate_indexer_namespace_for_testing()
        assert ns == "snow-test-19"

        ns = generate_indexer_namespace_for_testing('foo')
        assert ns == "snow-test-19"

    with override_environ(TRAVIS_JOB_ID=None, TEST_JOB_ID=None):
        ns = generate_indexer_namespace_for_testing('foo')
        assert re.match("foo-test-[0-9]+-", ns)


def test_dictionary_lookup():

    d1 = {'a': 3, 'b': 4}
    assert dictionary_lookup(d1, 'a') == 3
    assert dictionary_lookup(d1, 'b') == 4

    # TODO: These next few would be easier to do in unittest.TestCase where we could use self.assertRaisesRegexpe.
    #       Maybe a later version of pytest has the equivalent feature. -kmp 9-Feb-2020

    d2 = {'x': 10}
    try:
        dictionary_lookup(d2, 'y')
    except Exception as e:
        assert isinstance(e, DictionaryKeyError)
        assert str(e) == '''{'x': 10} has no 'y' key.'''
    else:
        pass  # XXX: functionality is broken with multiprocessing somehow -will 3/10/2020
        #raise AssertionError("No exception was raised where one was expected.")

    try:
        dictionary_lookup(17, 'z')
    except Exception as e:
        assert isinstance(e, DictionaryKeyError)
        assert str(e) == '''17 is not a dictionary.'''
    else:
        pass
        #raise AssertionError("No exception was raised where one was expected.")

    try:
        # String form of JSON isn't what's needed. It has to be parsed (i.e., a dict).
        dictionary_lookup(repr(d2), 'x')
    except Exception as e:
        assert isinstance(e, DictionaryKeyError)
        assert str(e) == '''"{'x': 10}" is not a dictionary.'''
    else:
        pass
        #raise AssertionError("No exception was raised where one was expected.")


@pytest.fixture
def simple_properties():
    return {
        'abc': 123,
        'animal': 'dog',
        'feline': 'cat',
        'names': {
            'will': 'dog',
            'bob': 'truck'
        }
    }


@pytest.fixture
def simple_calculated():
    return {
        'abcd': 1234
    }


@pytest.fixture
def complex_calculated():
    return {
        'abcd': 1234,
        'names': {
            'dog': 'will',
            'truck': 'bob'
        }
    }


@pytest.fixture
def failure_calculated1():
    return {
        'abc': 456
    }


@pytest.fixture
def failure_calculated2():
    return {
        'names': {
            'will': 'canine'
        }
    }


def check_original_props(props):
    """ Helper function that verifies correctness of the original fields, see fixture above """
    assert 'abc' in props
    assert 'animal' in props
    assert 'feline' in props
    assert 'names' in props
    assert 'will' in props['names']
    assert 'bob' in props['names']


def test_merge_calculated_into_properties(simple_properties, simple_calculated, complex_calculated,
                                          failure_calculated1, failure_calculated2):
    """ Tests some base cases with merging calculated into properties """
    # merge in one field - all should be the same
    props_copy = copy.deepcopy(simple_properties)
    merge_calculated_into_properties(props_copy, simple_calculated)
    check_original_props(props_copy)
    assert 'abcd' in props_copy

    # merge calculated sub-embedded fields
    props_copy = copy.deepcopy(simple_properties)
    merge_calculated_into_properties(props_copy, complex_calculated)
    check_original_props(props_copy)
    assert 'abcd' in props_copy
    assert 'dog' in props_copy['names']
    assert 'truck' in props_copy['names']

    # try to overwrite base field
    with pytest.raises(ValueError):
        props_copy = copy.deepcopy(simple_properties)
        merge_calculated_into_properties(props_copy, failure_calculated1)

    # try to override sub-embedded field
    with pytest.raises(ValueError):
        props_copy = copy.deepcopy(simple_properties)
        merge_calculated_into_properties(props_copy, failure_calculated2)


@pytest.fixture
def complex_properties():
    return {
        'abc': 123,
        'nested': [
            {
                'key': 'hello',
                'value': 'world'
            },
            {
                'key': 'dog',
                'value': 'cat'
            }
        ]
    }


@pytest.fixture
def calculated_list():
    return {
        'nested': [
            {
                'keyvalue': 'helloworld'
            },
            {
                'keyvalue': 'dogcat'
            }
        ]
    }


def test_merge_calculated_into_properties_array(complex_properties, calculated_list):
    """ Tests that we can do calculated properties on arrays of sub-embedded objects """
    props_copy = copy.deepcopy(complex_properties)
    merge_calculated_into_properties(props_copy, calculated_list)
    for entry in props_copy['nested']:
        for field in ['key', 'value', 'keyvalue']:
            assert field in entry


class TestCachedField:

    DEFAULT_TIMEOUT = 600

    @pytest.mark.flaky  # use of random.choice could generate collisions but extremely unlikely
    def test_cached_field_basic(self):
        def simple_update_function():
            return random.choice(range(10000))
        field = CachedField('simple1', update_function=simple_update_function)
        assert field.value is not None
        current = field.get()
        assert current == field.value
        assert field.get_updated() != current
        assert field.timeout == self.DEFAULT_TIMEOUT
        field.set_timeout(30)
        assert field.timeout == 30



from dcicutils.misc_utils import VirtualApp
from pyramid.paster import get_app


def make_standalone_app(*, config_file='development.ini', name='app'):
    """
    Creates an 'app' suitable for use in creating a testapp or VirtualApp.

    Args:
        name: a string naming the app (default 'app')
        config_file: a string naming the configuration file (default 'development.ini')

    Returns:

        An object, probably of type pyramid.router.Router, suitable for use in creating a
        webtest.TestApp or dcicutils.misc_utils.VirtualApp.

    """

    return get_app(config_file, name)


def make_dev_vapp(*, remote_user='TEST', environ=None, app=None, anonymous=False):
    """
    Creates a VirtualApp that, by default, simulates the TEST user of the development config,
    i.e., deployment on localhost (pserve development.ini).

    Args:
        remote_user: an email address or a reserved system token such as 'TEST' (the default).
        environ: a dictionary of environment bindings. The default is {"HTTP_ACCEPT": "application/json"}
        app: an app object to use instead of the default app newly generated on each call by by make_standalone_app()

    Returns:

        an object of type dcicutils.misc_utils.VirtualApp

    """

    environ = (environ if environ is not None else {'HTTP_ACCEPT': 'application/json'}).copy()

    if remote_user:
        environ['REMOTE_USER'] = remote_user

    app = app or make_standalone_app()
    return VirtualApp(app, environ)



import codecs
import collections
import io
import json
import uuid

from datetime import datetime
from dcicutils.misc_utils import ignored
from snovault.schema_validation import SerializingSchemaValidator
from jsonschema import Draft202012Validator
from jsonschema import RefResolver
from jsonschema.exceptions import ValidationError
import os
from pyramid.path import AssetResolver, caller_package
from pyramid.threadlocal import get_current_request
from pyramid.traversal import find_resource
from uuid import UUID
from .project_app import app_project
from .util import ensurelist

# This was originally an internal import from "." (__init__.py), but I have replaced that reference
# to avoid a circularity. No files should refer upward to "." for imports. -kmp 9-Jun-2020
from .resources import Item, COLLECTIONS


SERVER_DEFAULTS = {}


# TODO: Shouldn't this return func? Otherwise this:
#           @server_default
#           def foo(instance, subschema):
#               return ...something...
#       does (approximately):
#           SERVER_DEFAULTS['foo'] = lambda(instance, subschema): ...something...
#           server_default = None
#       It feels like the function should still get defined. -kmp 17-Feb-2023
def server_default(func):
    SERVER_DEFAULTS[func.__name__] = func


class NoRemoteResolver(RefResolver):
    def resolve_remote(self, uri):
        """ Resolves remote uri for files so we can cross reference across our own
            repos, which now contain base schemas we may want to use
        """
        if any(s in uri for s in ['http', 'https', 'ftp', 'sftp']):
            raise ValueError(f'Resolution disallowed for: {uri}')
        else:
            return load_schema(uri)


def favor_app_specific_schema(schema: str) -> str:
    """
    If the given schema refers to a schema (file) which exists in the app-specific schemas
    package/directory then favor that version of the file over the local version by returning
    a reference to that schema; otherwise just returns the given schema.

    For example, IF the given schema is snovault:access_key.json AND the current app is fourfront AND
    if the file encoded/schemas/access_key.json exists THEN returns: encoded:schemas/access_key.json

    This uses the dcicutils.project_utils mechanism to get the app-specific file/path name.
    """
    if isinstance(schema, str):
        schema_parts = schema.split(":")
        schema_project = schema_parts[0] if len(schema_parts) > 1 else None
        if schema_project != app_project().PACKAGE_NAME:
            schema_filename = schema_parts[1] if len(schema_parts) > 1 else schema_parts[0]
            app_specific_schema_filename = app_project().project_filename(f"/{schema_filename}")
            if os.path.exists(app_specific_schema_filename):
                schema = f"{app_project().PACKAGE_NAME}:{schema_filename}"
    return schema


def favor_app_specific_schema_ref(schema_ref: str) -> str:
    """
    If the given schema_ref refers to a schema (file) which exists in the app-specific schemas
    directory, AND it contains the specified element, then favor that version of the file over the
    local version by returning a reference to that schema; otherwise just returns the given schema_ref.

    For example, IF the given schema is mixins.json#/modified AND the current app is fourfront
    AND if the file encoded/schemas/mixins.json exists AND if that file contains the modified
    element THEN returns: file:///full-path-to/encoded/schemas/mixins.json#/modified

    This uses the dcicutils.project_utils mechanism to get the app-specific file/path name.
    """
    def json_file_contains_element(json_filename: str, json_element: str) -> bool:
        """
        If the given JSON file exists and contains the given JSON element name then
        returns True; otherwise returns False. The given JSON element may or may
        not begin with a slash. Currently only looks at one single top-level element.
        """
        if json_filename and json_element:
            try:
                with io.open(json_filename, "r") as json_f:
                    json_content = json.load(json_f)
                    json_element = json_element.strip("/")
                    if json_element:
                        if json_content.get(json_element):
                            return True
            except Exception:
                pass
        return False

    if isinstance(schema_ref, str):
        schema_parts = schema_ref.split("#")
        schema_filename = schema_parts[0]
        app_specific_schema_filename = app_project().project_filename(f"/schemas/{schema_filename}")
        if os.path.exists(app_specific_schema_filename):
            schema_element = schema_parts[1] if len(schema_parts) > 1 else None
            if schema_element:
                if json_file_contains_element(app_specific_schema_filename, schema_element):
                    schema_ref = f"file://{app_specific_schema_filename}#{schema_element}"
            else:
                schema_ref = f"file://{app_specific_schema_filename}"
    return schema_ref


def resolve_merge_ref(ref, resolver):
    with resolver.resolving(ref) as resolved:
        if not isinstance(resolved, dict):
            raise ValueError(
                f'Schema ref {ref} must resolve dict, not {type(resolved)}'
            )
        return resolved


def _update_resolved_data(resolved_data, value, resolver):
    # Assumes resolved value is dictionary.
    resolved_data.update(
        # Recurse here in case the resolved value has refs.
        resolve_merge_refs(
            # Actually get the ref value.
            resolve_merge_ref(value, resolver),
            resolver
        )
    )


def _handle_list_or_string_value(resolved_data, value, resolver):
    if isinstance(value, list):
        for v in value:
            _update_resolved_data(resolved_data, v, resolver)
    else:
        _update_resolved_data(resolved_data, value, resolver)


def resolve_merge_refs(data, resolver):
    if isinstance(data, dict):
        # Return copy.
        resolved_data = {}
        for k, v in data.items():
            if k == '$merge':
                _handle_list_or_string_value(resolved_data, v, resolver)
            else:
                resolved_data[k] = resolve_merge_refs(v, resolver)
    elif isinstance(data, list):
        # Return copy.
        resolved_data = [
            resolve_merge_refs(v, resolver)
            for v in data
        ]
    else:
        # Assumes we're only dealing with other JSON types
        # like string, number, boolean, null, not other
        # types like tuples, sets, functions, classes, etc.,
        # which would require a deep copy.
        resolved_data = data
    return resolved_data


def fill_in_schema_merge_refs(schema, resolver):
    """ Resolves $merge properties, custom $ref implementation from IGVF SNO2-6 """
    return resolve_merge_refs(schema, resolver)


def mixinSchemas(schema, resolver, key_name='properties'):
    mixinKeyName = 'mixin' + key_name.capitalize()
    mixins = schema.get(mixinKeyName)
    if mixins is None:
        return schema
    properties = collections.OrderedDict()
    bases = []
    for mixin in reversed(mixins):
        ref = mixin.get('$ref')
        if ref is not None:
            # For mixins check if there is an associated app-specific
            # schema file and favor that over the local one if any.
            # TODO: This may be controversial and up for discussion. 2023-05-27
            ref = favor_app_specific_schema_ref(ref)
            with resolver.resolving(ref) as resolved:
                mixin = resolved
        bases.append(mixin)
    for base in bases:
        for name, base_prop in base.items():
            prop = properties.setdefault(name, {})
            for k, v in base_prop.items():
                if k not in prop:
                    prop[k] = v
                    continue
                if prop[k] == v:
                    continue
                if key_name == 'facets':
                    continue  # Allow schema facets to override, as well.
                raise ValueError('Schema mixin conflict for %s/%s' % (name, k))
    # Allow schema properties to override
    base = schema.get(key_name, {})
    for name, base_prop in base.items():
        prop = properties.setdefault(name, {})
        for k, v in base_prop.items():
            prop[k] = v
    schema[key_name] = properties
    return schema


def linkTo(validator, linkTo, instance, schema):
    if not validator.is_type(instance, "string"):
        return

    request = get_current_request()
    collections = request.registry[COLLECTIONS]
    if validator.is_type(linkTo, "string"):
        base = collections.get(linkTo, request.root)
        linkTo = [linkTo] if linkTo else []
    elif validator.is_type(linkTo, "array"):
        base = request.root
    else:
        raise Exception("Bad schema")  # raise some sort of schema error

    try:
        item = find_resource(base, instance.replace(':', '%3A'))
    except KeyError:
        error = "%r not found" % instance
        yield ValidationError(error)
        return

    if not isinstance(item, Item):
        error = "%r is not a linkable resource" % instance
        yield ValidationError(error)
        return

    if linkTo and not set([item.type_info.name] + item.base_types).intersection(set(linkTo)):
        reprs = (repr(it) for it in linkTo)
        error = "%r is not of type %s" % (instance, ", ".join(reprs))
        yield ValidationError(error)
        return

    linkEnum = schema.get('linkEnum')
    if linkEnum is not None:
        if not validator.is_type(linkEnum, "array"):
            raise Exception("Bad schema")

        if not any(UUID(enum_uuid) == item.uuid for enum_uuid in linkEnum):
            reprs = ', '.join(repr(it) for it in linkTo)
            error = "%r is not one of %s" % (instance, reprs)
            yield ValidationError(error)
            return

    if schema.get('linkSubmitsFor'):
        userid = None
        for principal in request.effective_principals:
            if principal.startswith('userid.'):
                userid = principal[len('userid.'):]
                break

        if userid is not None:
            user = request.root[userid]
            submits_for = user.upgrade_properties().get('submits_for')
            if (submits_for is not None and
                    not any(UUID(uuid) == item.uuid for uuid in submits_for) and
                    not request.has_permission('submit_for_any')):
                error = "%r is not in user submits_for" % instance
                yield ValidationError(error)
                return


class IgnoreUnchanged(ValidationError):
    pass


def requestMethod(validator, requestMethod, instance, schema):
    ignored(instance, schema)
    if validator.is_type(requestMethod, "string"):
        requestMethod = [requestMethod]
    elif not validator.is_type(requestMethod, "array"):
        raise Exception("Bad schema")  # raise some sort of schema error
    request = get_current_request()
    if request.method not in requestMethod:
        reprs = ', '.join(repr(it) for it in requestMethod)
        error = "request method %r is not one of %s" % (request.method, reprs)
        yield IgnoreUnchanged(error)


def permission(validator, permission, instance, schema):
    ignored(instance, schema)
    if not validator.is_type(permission, "string"):
        raise Exception("Bad schema")  # raise some sort of schema error

    request = get_current_request()
    context = request.context
    if not request.has_permission(permission, context):
        error = "permission %r required" % permission
        yield IgnoreUnchanged(error)


VALIDATOR_REGISTRY = {}


def validators(validator, validators, instance, schema):
    if not validator.is_type(validators, "array"):
        raise Exception("Bad schema")  # raise some sort of schema error

    for validator_name in validators:
        validate = VALIDATOR_REGISTRY.get(validator_name)
        if validate is None:
            raise Exception('Validator %s not found' % validator_name)
        error = validate(instance, schema)
        if error:
            yield ValidationError(error)


def calculatedProperty(validator, linkTo, instance, schema):
    """ This is the validator for calculatedProperty - if we see this field on a submitted item
        we (normally) emit ValidationError since calculated properties cannot be submitted.

        However, if sub-embedded = True is set on the calculated property, allow submission
        if the schema is in one of two valid formats:
            1. It is an object field ie: calculated property on sub-embedded object
            2. It is an array of objects ie: calculated property applied across an array of
               sub-embedded objects.
    """
    ignored(instance)

    def schema_is_sub_embedded(schema):
        return schema.get('sub-embedded', False)

    def schema_is_object(schema):
        return schema.get('type') == 'object'

    def schema_is_array_of_objects(schema):
        return schema.get('type') == 'array' and schema_is_object(schema.get('items', {}))

    if (not schema_is_sub_embedded(schema)
            or (not schema_is_array_of_objects(schema) and not schema_is_object(schema))):
        yield ValidationError('submission of calculatedProperty disallowed')


class SchemaValidator(SerializingSchemaValidator):
    VALIDATORS = SerializingSchemaValidator.VALIDATORS.copy()
    VALIDATORS['calculatedProperty'] = calculatedProperty
    VALIDATORS['linkTo'] = linkTo
    VALIDATORS['permission'] = permission
    VALIDATORS['requestMethod'] = requestMethod
    VALIDATORS['validators'] = validators
    SERVER_DEFAULTS = SERVER_DEFAULTS


def load_schema(filename):
    filename = favor_app_specific_schema(filename)
    if isinstance(filename, dict):
        schema = filename
        resolver = NoRemoteResolver.from_schema(schema)
    else:
        utf8 = codecs.getreader("utf-8")
        asset = AssetResolver(caller_package()).resolve(filename)
        schema = json.load(utf8(asset.stream()),
                           object_pairs_hook=collections.OrderedDict)
        resolver = RefResolver('file://' + asset.abspath(), schema)
    # use mixinProperties, mixinFacets, mixinAggregations, and mixinColumns (if provided)
    schema = mixinSchemas(
        mixinSchemas(
            mixinSchemas(mixinSchemas(schema, resolver, 'properties'), resolver, 'facets'),
            resolver,
            'aggregations'
        ),
        resolver, 'columns'
    )
    schema = fill_in_schema_merge_refs(schema, resolver)

    # SchemaValidator is not thread safe for now
    SchemaValidator(schema, resolver=resolver)
    return schema


def extract_schema_default(schema, path):
    """ Extracts a schema default given a schema and an array path """
    if 'properties' in schema:
        schema = schema['properties']
    try:
        for key in path:
            obj = schema[key]
        return obj['default']
    except Exception:
        return None


def validate(schema, data, current=None, validate_current=False):
    """
    Validate the given data using a schema. Optionally provide current data
    to allow IgnoreUnchanged validation errors. If validate_current is set,
    will attempt to validate against current as well as data.
    current and validate_current should be used together if data is expected
    to have deleted fields that should be validated against.

    Args:
        schema (dict): item schema
        data (dict): new item contents
        current (dict): existing item contents
        validate_current (bool): whether to validate against current

    Returns:
        dict validated contents, list of errors
    """
    resolver = NoRemoteResolver.from_schema(schema)
    sv = SchemaValidator(schema, resolver=resolver, format_checker=Draft202012Validator.FORMAT_CHECKER)
    validated, errors = sv.serialize(data)
    # validate against current contents if validate_current is set
    if current and validate_current:
        validated_curr, errors_curr = sv.serialize(current)
        data_errors_detail = {str(err.path): err.message for err in errors}
        # add errors from validation of current if they are not in existing
        # error paths or already exist but have a different message
        for err in errors_curr:
            err_path = str(err.path)
            if err_path not in data_errors_detail:
                errors.append(err)
            elif data_errors_detail[err_path] != err.message:
                errors.append(err)
    filtered_errors = []
    for error in errors:
        # Possibly ignore validation if it results in no change to data
        if current is not None and isinstance(error, IgnoreUnchanged):
            # TODO: Should this next assignment be outside the 'for' and should this in loop be testing current_value?
            #       -kmp 13-Sep-2022
            current_value = current
            try:
                for key in error.path:
                    current_value = current_value[key]
            except Exception:
                current_value = None  # not found
            else:
                validated_value = validated
                try:
                    for key in error.path:
                        validated_value = validated_value[key]
                except Exception:
                    validated_value = None  # not found
                # TODO: Should this test be indented left by 4 spaces so that the other arms of the 'if' affect it?
                #       Right now those other arms set seemingly-unused variables. -kmp 7-Aug-2022
                if validated_value == current_value:
                    continue  # value is unchanged between data/current; ignore
        # Also ignore requestMethod and permission errors from defaults.
        if isinstance(error, IgnoreUnchanged):
            new_value = data
            try:
                for key in error.path:
                    # If it's in original data then either user passed it in
                    # or it's from PATCH object with unchanged data. If it's
                    # unchanged then it's already been skipped above.
                    new_value = new_value[key]
                    # XXX: the below I think creates a vulnerability where a user can revert
                    # a protected field to it's default?
                    # default = extract_schema_default(schema, error.path)
                    # if new_value == default:
                    #     continue
            except KeyError:
                # always True for us
                if current and validate_current:
                    for key in error.path:
                        val = current.get(key)
                    default = extract_schema_default(schema, error.path)
                    if val == default:
                        continue
                else:
                    continue
        filtered_errors.append(error)

    return validated, filtered_errors


def validate_request(schema, request, data=None, current=None):
    if data is None:
        data = request.json

    validated, errors = validate(schema, data, current)
    for error in errors:
        error_path = 'Schema: ' + '.'.join([str(p) for p in error.path])
        request.errors.add('body', error_path, error.message)

    if not errors:
        request.validated.update(validated)


def schema_validator(filename):
    schema = load_schema(filename)

    def validator(request):
        return validate_request(schema, request)

    return validator


def combine_schemas(a, b):
    if a == b:
        return a
    if not a:
        return b
    if not b:
        return a
    combined = collections.OrderedDict()
    for name in set(a.keys()).intersection(b.keys()):
        if a[name] == b[name]:
            combined[name] = a[name]
        elif name == 'type':
            combined[name] = sorted(set(ensurelist(a[name]) + ensurelist(b[name])))
        elif name == 'properties':
            combined[name] = collections.OrderedDict()
            for k in set(a[name].keys()).intersection(b[name].keys()):
                combined[name][k] = combine_schemas(a[name][k], b[name][k])
            for k in set(a[name].keys()).difference(b[name].keys()):
                combined[name][k] = a[name][k]
            for k in set(b[name].keys()).difference(a[name].keys()):
                combined[name][k] = b[name][k]
        elif name == 'items':
            combined[name] = combine_schemas(a[name], b[name])
        elif name in ('boost_values', 'facets'):
            combined[name] = collections.OrderedDict()
            combined[name].update(a[name])
            combined[name].update(b[name])
        elif name == 'columns':
            allValues = collections.OrderedDict()
            allValues.update(a[name])
            allValues.update(b[name])
            intersectedKeys = set(a[name].keys()).intersection(set(b[name].keys()))
            combined[name] = {k: v
                              for k, v in allValues.items()
                              if k in intersectedKeys}
    for name in set(a.keys()).difference(b.keys()):
        combined[name] = a[name]
    for name in set(b.keys()).difference(a.keys()):
        combined[name] = b[name]
    return combined


# for integrated tests
def utc_now_str():
    return datetime.utcnow().isoformat() + '+00:00'


@server_default
def userid(instance, subschema):  # args required by jsonschema-serialize-fork
    ignored(instance, subschema)
    return str(uuid.uuid4())


@server_default
def now(instance, subschema):  # args required by jsonschema-serialize-fork
    ignored(instance, subschema)
    return utc_now_str()



import re

from jsonschema import FormatChecker
from .server_defaults import (
    ACCESSION_PREFIX,
    ACCESSION_TEST_PREFIX,
)


# Codes we allow for testing go here.
ACCESSION_TEST_CODES = "BS|ES|EX|FI|FS|IN|SR|WF"

accession_re = re.compile(r'^%s[1-9A-Z]{9}$' % ACCESSION_PREFIX)

test_accession_re = re.compile(r'^%s(%s)[0-9]{4}([0-9][0-9][0-9]|[A-Z][A-Z][A-Z])$' % (
    ACCESSION_TEST_PREFIX, ACCESSION_TEST_CODES))

uuid_re = re.compile(r'(?i)[{]?(?:[0-9a-f]{4}-?){8}[}]?')


@FormatChecker.cls_checks("uuid")
def is_uuid(instance):
    # Python's UUID ignores all dashes, whereas Postgres is more strict
    # http://www.postgresql.org/docs/9.2/static/datatype-uuid.html
    return bool(uuid_re.match(instance))


def is_accession(instance):
    """Just a pattern checker."""
    # Unfortunately we cannot access the accessionType here
    return (
        accession_re.match(instance) is not None or
        test_accession_re.match(instance) is not None
    )



import argparse
import logging
import json
from pkg_resources import resource_filename

logger = logging.getLogger(__name__)
EPILOG = __doc__


def get_inserts(inserts_folder_name='inserts', inserts_file_name='workflow'):
    folder_name = resource_filename('encoded', 'tests/data/' + inserts_folder_name + '/')
    f_name = folder_name + inserts_file_name + '.json'
    with open(f_name) as f:
        items = json.loads(f.read())
    for insert_item in items:
        yield insert_item


def main():
    logging.basicConfig()
    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger('encoded').setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description="Run inserts through an upgrader", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--inserts-folder', help="Folder to use to get the file of inserts from. E.g. 'master-inserts' or 'inserts'. Defaults to 'inserts'.")
    parser.add_argument('item_type', help="Type of item or filename of inserts, in lowercase/schema-filename form, e.g. 'page', 'static_section'.")
    parser.add_argument('upgrader_method_name', help="Name of upgrader method to use as it is defined in upgrade/<item_type> folder, e.g. 'workflow_3_4'.")
    args = parser.parse_args()
    upgrader_module = __import__('encoded.upgrade.' + args.item_type, fromlist=[''])
    upgrader_fxn = getattr(upgrader_module, args.upgrader_method_name)
    results = []
    for item in get_inserts(args.inserts_folder or 'inserts', args.item_type):
        upgrader_fxn(item, None)  # Modifies in place
        results.append(item)
    print(json.dumps(results, indent=4, sort_keys=True))  # Return instead of print?


if __name__ == "__main__":
    main()



from dcicutils.misc_utils import environ_bool, PRINT
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.compat import is_nonstr_iter
from pyramid.interfaces import IAuthorizationPolicy
from pyramid.location import lineage
from zope.interface import implementer


DEBUG_PERMISSIONS = environ_bool("DEBUG_PERMISSIONS", default=False)


# This code (right now) is identical to that of https://github.com/lrowe/pyramid_localroles
# We are going to need to rip this out for CGAP. So first let's bring it in so we can debug and
# modify it while we implement a new permissions structure.

def local_principals(context, principals):
    """ The idea behind this is to process __ac_local_roles__ (and a boolean __ac_local_roles_block__
        to disable) and add local principals. This only works if you're in correct context, though,
        which does not seem to be the case.
    """
    local_principals = set()

    block = False
    for location in lineage(context):
        if block:
            break
        block = getattr(location, '__ac_local_roles_block__', False)
        local_roles = getattr(location, '__ac_local_roles__', None)

        if local_roles and callable(local_roles):
            local_roles = local_roles()

        if not local_roles:
            continue

        for principal in principals:
            try:
                roles = local_roles[principal]
            except KeyError:
                pass
            else:
                if not is_nonstr_iter(roles):
                    roles = [roles]
                local_principals.update(roles)

    if not local_principals:
        return principals

    local_principals.update(principals)

    if DEBUG_PERMISSIONS:
        PRINT("local_principals")
        PRINT(" context.collection=", context.collection)
        PRINT(" context.__acl__()=", context.__acl__())
        PRINT(" context.collection.__ac_local_roles_()=", context.__ac_local_roles__())
        PRINT("local_principals returning", local_principals)

    return local_principals


def merged_local_principals(context, principals):
    # XXX Possibly limit to prefix like 'role.'
    set_principals = frozenset(principals)
    local_principals = set()
    block = False
    for location in lineage(context):
        if block:
            break

        block = getattr(location, '__ac_local_roles_block__', False)
        local_roles = getattr(location, '__ac_local_roles__', None)

        if local_roles and callable(local_roles):
            local_roles = local_roles()

        if not local_roles:
            continue

        for principal, roles in local_roles.items():
            if not is_nonstr_iter(roles):
                roles = [roles]
            if not set_principals.isdisjoint(roles):
                local_principals.add(principal)

    if not local_principals:
        return principals

    local_principals.update(principals)

    local_principals = list(local_principals)

    if DEBUG_PERMISSIONS:
        PRINT("merged_local_principals")
        PRINT(" context.collection=", context.collection)
        PRINT(" context.__acl__()=", context.__acl__())
        PRINT(" context.collection.__ac_local_roles_()=", context.__ac_local_roles__())
        PRINT("merged_local_principals returning", local_principals)

    return local_principals


@implementer(IAuthorizationPolicy)
class LocalRolesAuthorizationPolicy(object):
    """Local roles authorization policy
    Objects may be given an ``__ac_local_roles__`` property which may be either a
    mapping or a callable that returns a mapping from principal id to a list of principals.
    """
    def __init__(self, wrapped_policy=None):
        if wrapped_policy is None:
            wrapped_policy = ACLAuthorizationPolicy()
        self.wrapped_policy = wrapped_policy

    def permits(self, context, principals, permission):
        principals = local_principals(context, principals)
        result = self.wrapped_policy.permits(context, principals, permission)
        if DEBUG_PERMISSIONS:
            PRINT("LocalRolesAuthorizationPolicy.permits")
            PRINT(" permission=", permission)
            PRINT(" principals=", principals)
            PRINT("LocalRolesAuthorizationPolicy.permits returning", result)
        return result

    def principals_allowed_by_permission(self, context, permission):
        principals = self.wrapped_policy.principals_allowed_by_permission(context, permission)
        result = merged_local_principals(context, principals)
        if DEBUG_PERMISSIONS:
            PRINT("LocalRolesAuthorizationPolicy.principals_allowed_by_permission")
            PRINT(" permission=", permission)
            PRINT(" principals=", principals)
            PRINT("LocalRolesAuthorizationPolicy.principals_allowed_by_permission returning", result)
        return result



from pyramid.settings import asbool


def parse_array(types, value):
    return [cast(types, v) for v in value.split(';') if v.strip()]


def parse_object(types, value):
    items = (part.split(':', 1) for part in value.split(',') if value.strip())
    return {k.strip(): cast(types, v) for k, v in items}


def parse_string(types, value):
    assert not types
    return value


def parse_ignore(types, value):
    return None


def parse_number(types, value):
    assert not types
    try:
        return int(value)
    except ValueError:
        return float(value)


def parse_integer(types, value):
    assert not types
    return int(value)


def parse_boolean(types, value):
    assert not types
    return asbool(value)


TYPE_BY_NAME = {
    'string': parse_string,
    'number': parse_number,
    'boolean': parse_boolean,
    'integer': parse_integer,
    'ignore': parse_ignore,
    'array': parse_array,
    'object': parse_object,
}


def cast(types, value):
    types = list(types) or ['string']
    type_name = types.pop()
    value = value.strip()
    if value.lower() == 'null':
        return None
    if value == '' and type_name != 'string':
        return None
    parse = TYPE_BY_NAME[type_name]
    return parse(types, value)


def convert(name, value):
    """ fieldname:<cast>[:<cast>...], value -> fieldname, cast(value)
    """
    parts = name.split(':')
    return parts[0], cast(parts[1:], value)


def cast_row_values(dictrows):
    """ Wrapper generator for typing csv.DictReader rows
    """
    for row in dictrows:
        yield dict(convert(name, value or '') for name, value in row.items())


def remove_nulls(dictrows):
    for row in dictrows:
        yield {
            name: value for name, value in row.items()
            if value is not None and name
        }



import argparse
import logging
import structlog

from dcicutils.env_utils import permit_load_data
from pyramid.paster import get_app
from pyramid.path import DottedNameResolver
from .. import configure_dbsession


log = structlog.getLogger(__name__)


EPILOG = __doc__


def load_data_should_proceed(env, allow_prod):
    """ Returns True on whether or not load_data should proceed.

    :param env: env we are on
    :param allow_prod: prod argument from argparse, defaults to False
    :return: True if load_data should continue, False otherwise
    """

    return permit_load_data(envname=env, allow_prod=allow_prod, orchestrated_app='cgap')  # TODO figure out app


def main(simulated_args=None):
    logging.basicConfig()
    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger('encoded').setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description="Load Test Data", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument('config_uri', help="path to configfile")
    parser.add_argument('--prod', action='store_true',
                        help="must be set to confirm this action is intended to happen on a production server")
    parser.add_argument('--overwrite', action='store_true',
                        help="must be set to update existing uuids with patch")
    parser.add_argument('--indir', help="directory to load items from")
    parser.add_argument('--itype', help="item type to load")
    args = parser.parse_args(simulated_args)

    # get the pyramids app
    app = get_app(args.config_uri, args.app_name)

    # create db schema
    configure_dbsession(app)

    env = app.registry.settings.get('env.name', '')

    allow_prod = args.prod
    load_data_by_type = DottedNameResolver().resolve("encoded.loadxl:load_data_by_type")

    if load_data_should_proceed(env, allow_prod):
        load_data_by_type(app, args.indir, args.overwrite, args.itype)


if __name__ == "__main__":
    main()



import re
import math
import itertools
import uuid
import structlog
from pyramid.view import view_config
from webob.multidict import MultiDict
from functools import reduce
from pyramid.httpexceptions import HTTPBadRequest
from urllib.parse import urlencode
from collections import OrderedDict
from copy import deepcopy
from snovault import (
    AbstractCollection,
    TYPES,
    COLLECTIONS,
    STORAGE
)
from snovault.elasticsearch import ELASTIC_SEARCH
from snovault.util import (
    debug_log,
)
from snovault.elasticsearch.indexer_utils import get_namespaced_index
from snovault.typeinfo import AbstractTypeInfo
from ..util import is_admin_request
from .lucene_builder import LuceneBuilder
from .search_utils import (
    find_nested_path, schema_for_field, get_es_index, get_es_mapping, is_date_field, is_numerical_field,
    is_array_of_numerical_field,
    execute_search, make_search_subreq, build_sort_dicts,
    NESTED, COMMON_EXCLUDED_URI_PARAMS, MAX_FACET_COUNTS,
)


log = structlog.getLogger(__name__)


def includeme(config):
    config.add_route('search', '/search{slash:/?}')
    config.scan(__name__)


sanitize_search_string_re = re.compile(r'[\\\+\-\&\|\!\(\)\{\}\[\]\^\~\:\/\\\*\?]')


class SearchBuilder:
    """ A monolithic object that encapsulates information needed to perform searches.
        The purpose of this class is to organize state + methods used for bootstrapping
        the search.

        Static methods in this class serve on of two purposes:
            1. Take state and use it to generate new state in the initializer.
            2. Functions that operate at the "leaf" and do not require state.

        The point is to split apart logic needed for query construction with logic needed for the
        API itself. Search is by far our most complicated API, thus there is a lot of state.
    """
    DEFAULT_SEARCH_FRAME = 'embedded'
    DEFAULT_HIDDEN = 'default_hidden'  # facet is hidden by default
    ADDITIONAL_FACETS = 'additional_facet'  # specifies an aggregation to compute in addition
    RESCUE_TERMS = 'rescue_terms'  # special facet field that contains terms that should always have buckets
    DEBUG = 'debug'  # search debug parameter
    CARDINALITY_RANGE = '-3.4028E38-*'
    PAGINATION_SIZE = 10  # for ECS, 10 is much better than 25, and may even do better when lowered
    MISSING = object()
    SEARCH_INFO_HEADER_TYPES = [
        'Workflow'  # TODO: add types here as needed
    ]

    def __init__(self, context, request, search_type=None, return_generator=False, forced_type='Search',
                 custom_aggregations=None, skip_bootstrap=False):
        self.context = context  # request context
        self.request = request  # request who requested a search
        self.response = {}

        # setup needed regardless of whether we are building from a base query or building a new one
        # from params
        self.types = self.request.registry[TYPES]  # all types in the system
        self.doc_types = self.set_doc_types(self.request, self.types, search_type)  # doc_types for this search
        self.es = self.request.registry[ELASTIC_SEARCH]  # handle to remote ES
        self.es_index = get_es_index(self.request, self.doc_types)  # what index we are searching on
        self.query = {}  # new search object, just raw lucene - NOT ElasticSearch-DSL

        # skip setup needed for building the query, if desired
        if not skip_bootstrap:
            self._bootstrap_query(search_type, return_generator, forced_type, custom_aggregations)

        # To be computed later, initialized to None here
        self.result = None
        self.from_ = None
        self.size = None
        self.facets = None
        self.search_session_id = None
        self.string_query = None
        self.facet_order_overrides = {}

    def _get_es_mapping_if_necessary(self):
        """ Looks in the registry to see if the single doc_type mapping is cached in the registry, which it
            should be - thus saving us some time from external API calls at the expense of application memory.
        """
        if len(self.doc_types) == 1:  # extract mapping from storage if we're searching on a single doc type
            item_type_snake_case = ''.join(['_' + c.lower() if c.isupper() else c for c in self.doc_types[0]]).lstrip('_')
            mappings = self.request.registry[STORAGE].read.mappings.get()
            if get_namespaced_index(self.request, item_type_snake_case) == self.es_index and self.es_index in mappings:
                return mappings[self.es_index]['mappings']['properties']
            else:  # new item was added after last cache update, get directly via API
                return get_es_mapping(self.es, self.es_index)
        return {}

    def _bootstrap_query(self, search_type=None, return_generator=False, forced_type='Search',
                         custom_aggregations=None):
        """ Helper method that will bootstrap metadata necessary for building a search query. """
        self.return_generator = return_generator  # whether or not this search should return a generator
        self.custom_aggregations = custom_aggregations  # any custom aggregations on this search
        self.forced_type = forced_type  # (mostly deprecated) search type
        self.principals = self.request.effective_principals  # permissions to apply to this search

        # Initialized via outside function call
        # schemas for doc_types
        self.schemas = [
            self.types[item_type].schema
            for item_type in self.doc_types
        ]
        # item_type hierarchy we are searching on
        self.search_types = self.build_search_types(self.types, self.doc_types) + [
            self.forced_type
        ]
        self.search_base = self.normalize_query(self.request, self.types, self.doc_types)
        self.search_frame = self.request.normalized_params.get('frame', self.DEFAULT_SEARCH_FRAME)  # embedded
        self.prepared_terms = self.prepare_search_term(self.request)
        self.additional_facets = self.request.normalized_params.getall(self.ADDITIONAL_FACETS)
        self.debug_is_active = self.request.normalized_params.getall(self.DEBUG)  # only used if admin
        self.source_fields = sorted(self.list_source_fields())

        # Can potentially make an outside API call, but ideally is cached
        # Only needed if searching on a single item type
        self.item_type_es_mapping = self._get_es_mapping_if_necessary()

    @property
    def forced_type_token(self):
        """ Do any processing needed to be applied to self.forced_type """
        return self.forced_type.lower()

    @classmethod
    def from_search(cls, context, request, s, return_generator=False):
        """ Builds a SearchBuilder object with a pre-built search by skipping the bootstrap
            initialization and setting self.query directly.

        :param context: context from request
        :param request: current request
        :param s: dictionary of lucene query
        :return: instance of this class with the search query dropped in
        """
        search_builder_instance = cls(context, request, return_generator=return_generator, skip_bootstrap=True)  # bypass (most of) bootstrap
        search_builder_instance.query = s  # parse compound query
        return search_builder_instance

    @staticmethod
    def build_search_types(types, doc_types):
        """ Builds search_types based on the given doc_types
        :param types: TypesTool from the registry
        :param doc_types: Type names we would like to search on
        :return: search_types, or a list of 'SearchResults' type candidates
        """
        search_types = []
        if len(doc_types) == 1:  # if we have one, add it and its base_type
            ti = types[doc_types[0]]
            search_types.append(ti.name + "SearchResults")
            if hasattr(ti, 'base_types'):
                for base_type in ti.base_types:
                    search_types.append(base_type + "SearchResults")

        # If we have more than one, compute and add common ancestors to search_types
        # TODO: handle more than 2 common ancestors
        else:
            base_types = []
            for ti in doc_types:
                if hasattr(types[ti], 'base_types'):
                    base_types.append(set(types[ti].base_types))
            common_ancestors = reduce(lambda x, y: x & y, base_types)
            if not common_ancestors:
                raise HTTPBadRequest("Tried to search on types with no common ancestor. This should never happen.")

            for ancestor in common_ancestors:
                if ancestor != "Item":
                    search_types.append(ancestor + "SearchResults")
            search_types.append("ItemSearchResults")
        return search_types

    @staticmethod
    def normalize_query(request, types, doc_types):
        """
        Normalize the query by calculating and setting request.normalized_params
        (a webob MultiDict) that is derived from custom query rules and also
        the list of doc_types specified by set_doc_types(). The normalize_param
        helper function finds field_schema for each query parameter and enforces
        a set of rules (see below). If the query item types differ from doc_types,
        override with doc_types

        :param request: the current Request
        :param types: registry[TYPES]
        :param doc_types: item_types to use for the search

        :returns: query string built from normalized params
        """

        # TODO: Optimize method structure here, see C4-71 PR comments -Will 6/24/2020
        def normalize_param(key, val):
            """
            Process each key/val in the original query param. As part of this,
            obtain the field schema for each parameter. Changes the query string
            to redirect the search to the normalized parameters
            Current rules:
            - for 'type', get name from types (from the registry)
            - append '.display_title' to any terminal linkTo query field
            - append '.display_title' to sorts on linkTo fields
            """
            # type param is a special case. use the name from TypeInfo
            if key == 'type' and val in types:
                return key, types[val].name

            # if key is sort, pass val as the key to this function
            # if it appends display title we know its a linkTo and
            # should be treated as such
            if key == 'sort':
                # do not use '-' if present
                sort_val = val[1:] if val.startswith('-') else val
                new_val, _ = normalize_param(sort_val, None)
                if new_val != sort_val:
                    val = val.replace(sort_val, new_val)
                return key, val

            # find schema for field parameter and drill down into arrays/subobjects
            field_schema = schema_for_field(key, request, doc_types)
            while field_schema and ('items' in field_schema or 'properties' in field_schema):
                try:
                    field_schema = field_schema['items']
                except KeyError:
                    pass
                try:
                    field_schema = field_schema['properties']
                except KeyError:
                    pass
            if field_schema and 'linkTo' in field_schema:
                # add display_title to terminal linkTo query fields
                if key.endswith('!'):  # handle NOT
                    return key[:-1] + '.display_title!', val
                return key + '.display_title', val
            else:
                return key, val

        # use a MultiDict to emulate request.params
        # TODO: Evaluate whether or not MultiDict is really useful here -Will 6/24/2020
        normalized_params = MultiDict(
            normalize_param(k, v)
            for k, v in request.params.items()
        )
        # overwrite 'type' if not equal to doc_types to ensure consistency
        if set(normalized_params.getall('type')) != set(doc_types):
            if 'type' in normalized_params:
                del normalized_params['type']
            for dtype in doc_types:
                normalized_params.add('type', dtype)

        # add the normalized params to the request
        # these will be used in place of request.params for the rest of search
        setattr(request, 'normalized_params', normalized_params)

        # the query string of the normalized search
        qs = '?' + urlencode([  # XXX: do we actually need to encode k,v  individually? -Will 6/24/2020
            (k.encode('utf-8'), v.encode('utf-8'))
            for k, v in request.normalized_params.items()
        ])
        return qs

    def prepare_search_term(self, request):
        """
        Prepares search terms by making a dictionary where the keys are fields and the values are arrays
        of query strings. This is an intermediary format which will be modified when constructing the
        actual search query.

        Ignore certain keywords, such as type, format, and field

        :param request: current request
        :return: dictionary mapping field --> query strings
        """
        prepared_terms = {}
        for field, val in request.normalized_params.items():
            if (field.startswith('validation_errors') or
                    field.startswith('aggregated_items') or
                    field == self.ADDITIONAL_FACETS):
                continue
            elif field == 'q':  # searched string has field 'q'
                # people shouldn't provide multiple queries, but if they do,
                # combine them with AND logic
                if 'q' in prepared_terms:
                    join_list = [prepared_terms['q'], val]
                    prepared_terms['q'] = ' AND '.join(join_list)
                else:
                    prepared_terms['q'] = val
            elif field not in COMMON_EXCLUDED_URI_PARAMS + ['type']:
                if 'embedded.' + field not in prepared_terms:
                    prepared_terms['embedded.' + field] = []
                prepared_terms['embedded.' + field].append(val)
        return prepared_terms

    @staticmethod
    def set_doc_types(request, types, search_type):
        """
        Set the type of documents resulting from the search; order and check for
        invalid types as well. If a forced search_type is enforced, use that;
        otherwise, set types from the query params. Default to Item if none set.

        :param request: the current Request
        :param types: registry[TYPES]
        :param search_type: forced search item type

        :returns: list: the string item types to use for the search
        :raises: HTTPBadRequest: if an invalid item type is supplied
        """
        if search_type is None:
            doc_types = request.params.getall('type')
            if '*' in doc_types:
                doc_types = ['Item']
        else:
            doc_types = [search_type]
        # Normalize to item_type
        try:
            doc_types = sorted({types[name].name for name in doc_types})
        except KeyError:
            # Check for invalid types
            bad_types = [t for t in doc_types if t not in types]
            msg = "Invalid type: {}".format(', '.join(bad_types))
            raise HTTPBadRequest(explanation=msg)
        if len(doc_types) == 0:
            doc_types = ['Item']
        return doc_types

    def add_search_header_if_needed(self):
        """
        Get static section (if applicable) when searching a single item type
        Note: Because we rely on 'source', if the static_section hasn't been indexed
        into Elasticsearch it will not be loaded

        Only check for this if the item type is declared to have one in
        the class constant SEARCH_INFO_HEADER_TYPES
        """
        if (len(self.doc_types) == 1 and 'Item' not in self.doc_types and
                self.doc_types[0] in self.SEARCH_INFO_HEADER_TYPES):
            search_term = 'search-info-header.' + self.doc_types[0]
            # XXX: this could be cached application side as well
            try:
                static_section = self.request.registry['collections']['StaticSection'].get(search_term)
            except Exception:  # NotFoundError not caught, search could fail
                static_section = None
            if static_section and hasattr(static_section.model, 'source'):  # extract from ES structure
                item = static_section.model.source['object']
                self.response['search_header'] = {}
                self.response['search_header']['content'] = item.get('content', 'Content Missing')
                self.response['search_header']['title'] = item.get('title', item['display_title'])
                self.response['search_header']['filetype'] = item.get('filetype', 'No filetype')
            elif static_section and hasattr(static_section.model, 'data'):  # extract form DB structure
                item = static_section.upgrade_properties()
                self.response['search_header'] = {}
                self.response['search_header']['content'] = item.get('body', 'Content Missing')
                self.response['search_header']['title'] = item.get('title', 'No title')
                self.response['search_header']['filetype'] = item.get('filetype', 'No filetype')
            else:
                pass  # no static header found

    def set_pagination(self):
        """
        Fill from_ and size parameters for search if given in the query string
        """
        from_ = self.request.normalized_params.get('from', 0)
        size = self.request.normalized_params.get('limit', self.PAGINATION_SIZE)
        if size in ('all', ''):
            size = "all"
        else:
            try:
                size = int(size)
            except ValueError:
                size = self.PAGINATION_SIZE
            try:
                from_ = int(from_)
            except ValueError:
                from_ = 0
        self.from_, self.size = from_, size

    def build_type_filters(self):
        """
        Set the type filters for the search. If no doc_types, default to Item.
        This also sets the facet filter override, allowing you to apply custom facet ordering
        by specifying the FACET_ORDER_OVERRIDE field on the type definition. See VariantSample
        or _sort_custom_facets for examples.
        """
        if not self.doc_types:
            self.doc_types = ['Item']
        else:
            for item_type in self.doc_types:
                ti = self.types[item_type]
                if hasattr(ti, 'factory'):  # if not abstract
                    self.facet_order_overrides.update(getattr(ti.factory, 'FACET_ORDER_OVERRIDE', {}))

                qs = urlencode([
                    (k.encode('utf-8'), v.encode('utf-8'))
                    for k, v in self.request.normalized_params.items()
                    if k != "limit" and k != "from" and not (k == 'type' and self.types.all.get('Item' if v == '*' else v) is ti)
                ])
                self.response['filters'].append({
                    'field': 'type',
                    'term': ti.name,
                    'remove': '{}?{}'.format(self.request.path, qs)
                })

    def clear_filters_setup(self):
        """
        Clear Filters URI path

        Make a URI path that clears all non-datatype filters
        and leaves in `q` (search query) params, if present.
        Also preserves currentAction=selection, if is set.

        :returns: A URL path
        """
        seach_query_specs = self.request.normalized_params.getall('q')
        seach_query_url = urlencode([("q", seach_query) for seach_query in seach_query_specs])
        # types_url will always be present (always >=1 doc_type)
        types_url = urlencode([("type", typ) for typ in self.doc_types])
        current_action = self.request.normalized_params.get('currentAction')

        clear_qs = types_url or ''
        if seach_query_url:
            clear_qs += '&' + seach_query_url
        if current_action == 'selection':
            clear_qs += '&currentAction=selection'
        current_search_sort = self.request.normalized_params.getall('sort')
        current_search_sort_url = urlencode([("sort", s) for s in current_search_sort])
        if current_search_sort_url:
            clear_qs += '&' + current_search_sort_url
        return self.request.route_path(self.forced_type_token, slash='/') + (('?' + clear_qs) if clear_qs else '')

    def initialize_search_response(self):
        """ Initializes the search response """
        self.response = {
            '@context': self.request.route_path('jsonld_context'),
            '@id': '/' + self.forced_type_token + '/' + self.search_base,
            '@type': self.search_types,
            'title': self.forced_type,
            'filters': [],
            'facets': [],
            '@graph': [],
            'notification': '',
            'sort': {},
            'clear_filters': self.clear_filters_setup()
        }
        self.add_search_header_if_needed()
        self.set_pagination()
        self.build_type_filters()

    def list_source_fields(self):
        """
        Returns set of fields that are requested by user or default fields.
        These fields are used to further limit the results from the search.
        Note that you must provide the full fieldname with embeds, such as:
        'field=biosample.biosource.individual.organism.name' and not just
        'field=name'
        """
        fields_requested = self.request.normalized_params.getall('field')
        if fields_requested:
            fields = ['embedded.@id', 'embedded.@type']
            for field in fields_requested:
                fields.append('embedded.' + field)
        elif self.search_frame == 'embedded':
            fields = [self.search_frame + '.*']
        elif self.search_frame in ['object', 'raw']:
            # frame=raw corresponds to 'properties' in ES
            if self.search_frame == 'raw':
                frame = 'properties'
            else:
                frame = self.search_frame
            # let embedded be searched as well (for faceting)
            fields = ['embedded.*', frame + '.*']
        else:
            fields = ['embedded.*']
        return fields

    def build_query(self):
        """
        Bootstraps our query format, building the q= part of the query if one is specified.
        If multiple are specified the first one that occurs in the URL will be used.
        """
        query_info = {}
        string_query = None
        query_dict = {'query': {'bool': {}}}
        # handle field, frame
        self.query['_source'] = self.source_fields
        # locate for 'q' query, if any
        for field, value in self.prepared_terms.items():
            if field == 'q':
                query_info['query'] = value
                query_info['lenient'] = True
                query_info['default_operator'] = 'AND'
                query_info['fields'] = ['full_text']
                string_query = {'must': {'simple_query_string': query_info}}
                query_dict = {'query': {'bool': string_query}}
                break
        self.query.update(query_dict)
        self.string_query = string_query

    def set_sort_order(self):
        """
        sets sort order for elasticsearch results
        example: /search/?type=Biosource&sort=display_title
        will sort by display_title in ascending order. To set descending order,
        use the "-" flag: sort_by=-date_created.
        Sorting is done alphatbetically, case sensitive by default.

        ES5: simply pass in the sort OrderedDict into search.sort
        """

        # Prefer sort order specified in request, if any
        requested_sorts = self.request.normalized_params.getall('sort')
        text_search = self.prepared_terms.get('q')
        sort, result_sort = build_sort_dicts(requested_sorts, self.request, self.doc_types, text_search)

        # Otherwise we use a default sort only when there's no text search to be ranked
        if not sort and text_search and text_search != '*':
            # Multi-level sort. See http://www.elastic.co/guide/en/elasticsearch/guide/current/_sorting.html#_multilevel_sorting & https://stackoverflow.com/questions/46458803/python-elasticsearch-dsl-sorting-with-multiple-fields
            self.query['sort'] = [{'_score': {"order": "desc"}},
                                  {'embedded.date_created.raw': {'order': 'desc', 'unmapped_type': 'keyword'},
                                   'embedded.label.raw': {'order': 'asc', 'unmapped_type': 'keyword', 'missing': '_last'}},
                                  {'_id': {'order': 'asc'}}  # ES7 - _uid removed, now use _id
                ]
                # 'embedded.uuid.raw' (instd of _id) sometimes results in 400 bad request : 'org.elasticsearch.index.query.QueryShardException: No mapping found for [embedded.uuid.raw] in order to sort on'

            self.response['sort'] = result_sort = {'_score': {"order": "desc"}}

        if sort and result_sort:
            self.response['sort'] = result_sort
            self.query['sort'] = sort

    def _initialize_additional_facets(self, facets_so_far, current_type_schema):
        """ Helper function for below method that handles additional_facets URL param

            :param facets_so_far: list to add additional_facets to
            :param current_type_schema: schema of the item we are faceting on
        """
        for extra_facet in self.additional_facets:
            aggregation_type = 'terms'  # default

            # determine if nested
            if self.item_type_es_mapping and find_nested_path(extra_facet, self.item_type_es_mapping):
                aggregation_type = 'nested'  # handle nested

            # check if defined in facets
            if 'facets' in current_type_schema:
                schema_facets = current_type_schema['facets']
                if extra_facet in schema_facets:
                    if not schema_facets[extra_facet].get('disabled', False):
                        facets_so_far.append((extra_facet, schema_facets[extra_facet]))
                    continue  # if we found the facet, always continue from here

            # not specified as facet - infer range vs. term based on schema
            field_definition = schema_for_field(extra_facet, self.request, self.doc_types)
            if not field_definition:  # if not on schema, try "terms"
                facets_so_far.append((
                    extra_facet, {'title': extra_facet.title()}
                ))
            else:
                t = field_definition.get('type', None)
                if not t:
                    log.error('Encountered an additional facet that has no type! %s' % field_definition)
                    continue  # drop this facet

                # terms for string
                if t == 'string':
                    facets_so_far.append((
                        extra_facet, {'title': extra_facet.title(), 'aggregation_type': aggregation_type}
                    ))
                else:  # try stats
                    aggregation_type = 'stats'
                    facets_so_far.append((
                        extra_facet, {
                            'title': field_definition.get('title', extra_facet.title()),
                            'aggregation_type': aggregation_type,
                            'number_step': 'any'
                        }
                    ))

    def initialize_facets(self):
        """
        Initialize the facets used for the search. If searching across multiple
        doc_types, only use the default 'Data Type' and 'Status' facets.
        Add facets for custom url filters whether or not they're in the schema.
        TODO: clean up this method - see comments in C4-71 PR

        :param doc_types: Item types (@type) for which we are performing a search for.
        :param prepared_terms: terms to match in ES, keyed by ES field name.
        :param schemas: List of OrderedDicts of schemas for doc_types.

        :returns: list: tuples containing (0) ElasticSearch-formatted field name (e.g. `embedded.status`)
                        and (1) list of terms for it.
        """
        if len(self.doc_types) > 1:  # only provide this if we are searching on more than one type
            facets = [
                # More facets will be appended to this list from item schema plus from any currently-active filters (as requested in URI params).
                ('type', {'title': 'Data Type'})
            ]
        else:
            facets = []

        append_facets = [
            # Facets which will be appended after those which are in & added to `facets`
            # ('status', {'title': 'Status'}), XXX: uncomment this if you want status facet

            # TODO: Re-enable below line if/when 'range' URI param queries for date & numerical fields are implemented.
            # ('date_created', {'title': 'Date Created', 'hide_from_view' : True, 'aggregation_type' : 'date_histogram' })
        ]
        validation_error_facets = [
            ('validation_errors.name', { 'title': 'Validation Errors', 'order': 999 })
        ]
        current_type_schema = self.request.registry[TYPES][self.doc_types[0]].schema
        self._initialize_additional_facets(append_facets, current_type_schema)
        # hold disabled facets from schema; we also want to remove these from the prepared_terms facets
        disabled_facet_fields = set()

        # Add facets from schema if one Item type is defined.
        # Also, conditionally add extra appendable facets if relevant for type from schema.
        if len(self.doc_types) == 1 and self.doc_types[0] != 'Item':
            if 'facets' in current_type_schema:
                schema_facets = OrderedDict(current_type_schema['facets'])
                for schema_facet in schema_facets.items():
                    if schema_facet[1].get('disabled', False) or schema_facet[1].get(self.DEFAULT_HIDDEN, False):
                        disabled_facet_fields.add(schema_facet[0])
                        continue  # Skip disabled facets.
                    facets.append(schema_facet)

        # Add facets for any non-schema ?field=value filters requested in the search (unless already set, via used_facet_fields)
        used_facet_fields = set()
        used_facet_titles = set()
        for facet in facets + append_facets:
            used_facet_fields.add(facet[0])
            if 'title' in facet[1]:
                used_facet_titles.add(facet[1]['title'])

        for field in self.prepared_terms:
            if field.startswith('embedded'):

                # Will become, e.g. ['embedded', 'experiments_in_set', 'files', 'file_size', 'from']
                split_field = field.strip().split('.')
                use_field = '.'.join(split_field[1:]) # e.g. "experiments_in_set.files.file_size.from"

                if use_field in used_facet_fields or use_field in disabled_facet_fields:
                    # Cancel if already in facets or is disabled (first check, before more broad check re: agg_type:stats, etc)
                    continue

                # Use the last part of the split field to get the field title
                title_field = split_field[-1]

                # workaround: if query has a '!=' condition, title_field ends with '!'. This prevents to find the proper display title.
                # TODO: instead of workaround, '!' could be excluded while generating query results
                if use_field.endswith('!'):
                    use_field = use_field[:-1]
                    title_field = title_field[:-1]

                # if searching for a display_title, use the title of parent object
                # use `is_object_title` to keep track of this
                if title_field == 'display_title' and len(split_field) > 1:
                    title_field = split_field[-2]
                    is_object_title = True
                else:
                    is_object_title = False


                # 'terms' is the default per-term bucket aggregation for all non-schema facets
                if self.item_type_es_mapping and find_nested_path(field, self.item_type_es_mapping):
                    aggregation_type = 'nested'
                else:
                    aggregation_type = 'terms'

                # If we have a range filter in the URL, strip out the ".to" and ".from"
                if title_field == 'from' or title_field == 'to':
                    if len(split_field) >= 3:
                        f_field = ".".join(split_field[1:-1])
                        field_schema = schema_for_field(f_field, self.request, self.doc_types)

                        if field_schema:
                            # field could be a date, numerical, or array of numerical
                            if (is_date_field(field, field_schema) or
                                    is_numerical_field(field_schema) or
                                    is_array_of_numerical_field(field_schema)):
                                title_field = field_schema.get("title", f_field)
                                use_field = f_field
                                aggregation_type = 'stats'

                # At moment is equivalent to `if aggregation_type == 'stats'` until/unless more agg types are added for _facets_.
                if aggregation_type == 'stats':
                    # Remove completely if duplicate (e.g. don't need to have` .from` and `.to` both present)
                    if use_field in used_facet_fields or use_field in disabled_facet_fields:
                        continue
                    # Facet would be otherwise added twice if both `.from` and `.to` are requested.

                for schema in self.schemas:
                    if title_field in schema['properties']:
                        title_field = schema['properties'][title_field].get('title', title_field)
                        # see if the title field conflicts for is_object_title facets
                        if is_object_title and title_field in used_facet_titles:
                            title_field += ' (Title)'
                        break

                used_facet_fields.add(use_field)
                facets.append((
                    use_field,
                    { 'title': title_field, 'aggregation_type': aggregation_type }
                ))

        # Append additional facets (status, validation_errors, ...) at the end of
        # list unless were already added via schemas, etc.
        used_facet_fields = { facet[0] for facet in facets } # Reset this
        for ap_facet in append_facets + validation_error_facets:
            if ap_facet[0] not in used_facet_fields:
                used_facet_fields.add(ap_facet[0])
                facets.append(ap_facet)
        return facets

    def assure_session_id(self):
        """ Add searchSessionID information if not part of a sub-request, a generator or a limit=all search """
        if (
            self.request.__parent__ is None and
            not getattr(self, "return_generator", None) and
            getattr(self, "size", 25) != "all"
        ):  # Probably unnecessary, but skip for non-paged, sub-reqs, etc.
            self.search_session_id = self.request.cookies.get('searchSessionID', 'SESSION-' + str(uuid.uuid1()))

    def build_search_query(self):
        """ Builds the search query utilizing a combination of helper methods within this class
            to build intermediary structures and LuceneBuilder function calls to handle building
            the actual Elasticsearch query.
        """
        self.build_query()
        self.set_sort_order()

        # Transform into filtered search
        self.query, query_filters = LuceneBuilder.build_filters(self.request, self.query, self.response,
                                                                 self.principals, self.doc_types,
                                                                 self.item_type_es_mapping)
        # Prepare facets in intermediary structure
        self.facets = self.initialize_facets()

        # Transform filter search into filter + faceted search
        self.query = LuceneBuilder.build_facets(self.query, self.facets, query_filters, self.string_query,
                                                self.request, self.doc_types, self.custom_aggregations, self.size,
                                                self.from_, self.item_type_es_mapping)

        # Add preference from session, if available
        # This just sets the value on the class - it is passed to execute_search later
        self.assure_session_id()

    @staticmethod
    def fix_and_replace_nested_doc_count(result_facet, aggregations, full_agg_name):
        """
        3 things must happen here (all occurring by side-effect, not value):
            1. front-end does not care about 'nested', only what the inner thing is, so lets pretend (so it doesn't break)
            2. We must overwrite the "second level" doc_count with the "third level" because the "third level"
               is the 'root' level doc_count, which is what we care about, NOT the nested doc count
            3. We must then re-sort the aggregations so they show up in from greatest to least doc_count wrt the root
               level count instead of the "old" nested doc count.

        :param result_facet: facet to be created - 'aggregation_type' is overwritten as 'terms'
        :param aggregations: handle to all aggregations that we can access based on name
        :param full_agg_name: full name of the aggregation
        """
        result_facet['aggregation_type'] = 'terms'
        term_to_bucket = {}  # so we can deduplicate keys
        source_aggregation = aggregations[full_agg_name]['primary_agg']
        primary_buckets = source_aggregation['primary_agg']['buckets']
        for bucket in primary_buckets:
            if 'primary_agg_reverse_nested' in bucket:
                bucket['doc_count'] = bucket['primary_agg_reverse_nested']['doc_count']
            if bucket['key'] not in term_to_bucket:
                term_to_bucket[bucket['key']] = bucket
        if 'requested_agg' in source_aggregation:
            requested_buckets = source_aggregation['requested_agg']['buckets']
            for bucket in requested_buckets:
                if 'primary_agg_reverse_nested' in bucket:
                    bucket['doc_count'] = bucket['primary_agg_reverse_nested']['doc_count']
                if bucket['key'] not in term_to_bucket:
                    term_to_bucket[bucket['key']] = bucket

        result_facet['terms'] = sorted(list(term_to_bucket.values()),key=lambda d: d['primary_agg_reverse_nested']['doc_count'], reverse=True)

    def format_facets(self, es_results):
        """
        This method processes the 'aggregations' component of the ES response.
        It does this by creating result_facet frames for all facets retrieved from ES
        and populating the frame with the relevant aggregation info depending on it's type.

        Format the facets for the final results based on the es results.
        Sort based off of the 'order' of the facets
        These are stored within 'aggregations' of the result.

        If the frame for the search != embedded, return no facets
        """
        result = []
        if self.search_frame != 'embedded':
            return result

        # Loading facets in to the results
        if 'aggregations' not in es_results:
            return result

        aggregations = es_results['aggregations']['all_items']

        # Sort facets by order (ascending).
        # If no order is provided, assume 0 to
        # retain order of non-explicitly ordered facets
        for field, facet in sorted(self.facets, key=lambda fct: fct[1].get('order', 10000)):
            if facet.get(self.DEFAULT_HIDDEN, False) and field not in self.additional_facets:  # skip if specified
                continue

            # Build result frame for the front-end
            result_facet = {
                'field': field,
                'title': facet.get('title', field),
                'total': 0
                # To be added depending on facet['aggregation_type']: 'terms', 'min', 'max', 'min_as_string', 'max_as_string', ...
            }

            result_facet.update({k: v for k, v in facet.items() if k not in result_facet.keys()})
            field_agg_name = field.replace('.', '-')
            full_agg_name = facet['aggregation_type'] + ':' + field_agg_name

            if full_agg_name in aggregations:  # found an agg for this field

                # process stats agg
                if facet['aggregation_type'] == 'stats':
                    result_facet['total'] = aggregations[full_agg_name]['doc_count']
                    # Used for fields on which can do range filter on, to provide min + max bounds
                    for k in aggregations[full_agg_name]['primary_agg'].keys():
                        result_facet[k] = aggregations[full_agg_name]['primary_agg'][k]

                # nested stats aggregations have a second "layer" for reverse_nested
                elif facet['aggregation_type'] == 'nested:stats':
                    result_facet['total'] = aggregations[full_agg_name]['primary_agg']['doc_count']
                    for k in aggregations[full_agg_name]['primary_agg']['primary_agg'].keys():
                        result_facet[k] = aggregations[full_agg_name]['primary_agg']['primary_agg'][k]

                # process range agg
                elif facet['aggregation_type'] in ['range', 'nested:range']:
                    # Shift the bucket location
                    bucket_location = aggregations[full_agg_name]['primary_agg']
                    if 'buckets' not in bucket_location:  # account for nested structure
                        bucket_location = bucket_location['primary_agg']

                    # TODO - refactor ?
                    # merge bucket labels from ranges into buckets
                    for r in result_facet['ranges']:
                        for b in bucket_location['buckets']:
                            if (r.get('from', self.MISSING) == b.get('from', self.MISSING) and
                                    r.get('to', self.MISSING) == b.get('to', self.MISSING)):
                                r['doc_count'] = b['doc_count']
                                break

                # process terms agg
                else:
                    # do the below, except account for nested agg structure
                    if facet['aggregation_type'] == NESTED:
                        self.fix_and_replace_nested_doc_count(result_facet, aggregations, full_agg_name)
                        result.append(result_facet)
                        continue

                    def extract_buckets(path):
                        if 'buckets' not in path:
                            path = path['primary_agg']
                        if 'buckets' not in path:
                            raise Exception('No buckets found on terms agg!')
                        return path['buckets']

                    term_to_bucket = {}  # so we can deduplicate keys
                    source_aggregation = aggregations[full_agg_name]
                    if 'requested_agg' in source_aggregation:
                        for bucket in extract_buckets(source_aggregation['requested_agg']):
                            if bucket['key'] not in term_to_bucket:
                                term_to_bucket[bucket['key']] = bucket

                    # always present
                    for bucket in extract_buckets(source_aggregation['primary_agg']):
                        if bucket['key'] not in term_to_bucket:
                            term_to_bucket[bucket['key']] = bucket

                    # bring in rescue terms
                    if self.RESCUE_TERMS in facet:
                        for rescue_term in facet[self.RESCUE_TERMS]:
                            if rescue_term not in term_to_bucket:
                                term_to_bucket[rescue_term] = {
                                    'key': rescue_term,
                                    'doc_count': 0
                                }

                    result_facet['terms'] = list(term_to_bucket.values())

                # XXX: not clear this functions as intended - Will 2/17/2020
                if len(aggregations[full_agg_name].keys()) > 2:
                    result_facet['extra_aggs'] = {k: v for k, v in aggregations[full_agg_name].items() if
                                                  k not in ('doc_count', 'primary_agg')}

            # Minor improvement for UI/UX -- if no 'validation errors' facet is included
            # but only 1 "No value" term, then exclude the facet so is not shown
            if field == "validation_errors.name":
                validation_errors_terms_len = len(result_facet["terms"])
                if (
                    validation_errors_terms_len == 0 or
                    (validation_errors_terms_len == 1 and result_facet["terms"][0]["key"] == "No value")
                ):
                    continue

            result.append(result_facet)

        # TODO ALEX: Client will reject 'nested:stats' so overwritten here.
        #            Ideally, the client should accept 'stats', 'terms', 'nested:terms', 'nested:stats',
        #            and just treat the nested aggs exactly the same.
        for facet in result:
            agg_type = facet.get("aggregation_type")
            override = None
            if agg_type == 'nested:stats':
                override = 'stats'
            elif agg_type == 'nested:range':
                override = 'range'

            # apply override
            if override is not None:
                facet["aggregation_type"] = override

        return result

    @staticmethod
    def format_extra_aggregations(es_results):
        """
        Extracts any extra aggregations results returned from elasticsearch

        :param es_results: dictionary response from es
        :return: dictionary mapping field -> agg_value, varies based on type of aggregation
        """
        if 'aggregations' not in es_results:
            return {}
        return {k: v for k, v in es_results['aggregations'].items()
                if k != 'all_items'}

    def get_collection_actions(self):
        """
        Use this method to see actions available on an item type (collection) in the request context

        :return: actions available for this collection at this time
        """
        type_info = self.types[self.doc_types[0]]  # only care about the "first" collection
        collection = self.request.registry[COLLECTIONS].get(type_info.name)
        if hasattr(collection, 'actions'):
            return collection.actions(self.request)
        else:
            return None

    @staticmethod
    def build_initial_columns(used_type_schemas):

        columns = OrderedDict()

        # Add title column, at beginning always
        columns['display_title'] = {
            "title": "Title",
            "order": -1000
        }

        for schema in used_type_schemas:
            if 'columns' in schema:
                schema_columns = OrderedDict(schema['columns'])
                # Add all columns defined in schema
                for name, obj in schema_columns.items():
                    if name not in columns:
                        columns[name] = obj
                    else:
                        # If @type or display_title etc. column defined in schema, then override defaults.
                        columns[name].update(schema_columns[name])

        # Add status column, if not present, at end.
        if 'status' not in columns:
            columns['status'] = {
                "title": "Status",
                "default_hidden": True,
                "order": 980
            }

        # Add date column, if not present, at end.
        if 'date_created' not in columns:
            columns['date_created'] = {
                "title": "Date Created",
                "colTitle": "Created",
                "default_hidden": True,
                "order": 1000
            }

        return columns

    def build_table_columns(self):
        """ Constructs an ordered dictionary of column information to be rendered by
            the front-end. If this functionality is needed outside of general search, this
            method should be moved to search_utils.py.
        """

        columns = SearchBuilder.build_initial_columns(self.schemas)

        if self.request.normalized_params.get('currentAction') in ('selection', 'multiselect'):
            return columns

        any_abstract_types = 'Item' in self.doc_types
        if not any_abstract_types:  # Check explictly-defined types to see if any are abstract.
            type_infos = [self.request.registry[TYPES][t] for t in self.doc_types if t != 'Item']
            for ti in type_infos:
                # We use `type` instead of `isinstance` since we don't want to catch subclasses.
                if type(ti) is AbstractTypeInfo:
                    any_abstract_types = True
                    break

        # Add type column if any abstract types in search
        if any_abstract_types:
            columns['@type'] = {
                "title": "Item Type",
                "colTitle": "Type",
                "order": -980,
                "description": "Type or category of Item",
                # Alternative below, if we want type column to be available but hidden by default in selection mode:
                # "default_hidden": request.normalized_params.get('currentAction') == 'selection'
            }

        return columns

    def _format_results(self, hits):
        """
        Loads results to pass onto UI
        Will retrieve the desired frame from the search hits and automatically
        add 'validation_errors' and 'aggregated_items' frames if they are present
        """
        fields_requested = self.request.normalized_params.getall('field')
        if fields_requested:
            frame = 'embedded'
        elif self.search_frame:
            frame = self.search_frame
        else:
            frame = 'embedded'

        if frame in ['embedded', 'object', 'raw']:
            # transform 'raw' to 'properties', which is what is stored in ES
            if frame == 'raw':
                frame = 'properties'
            for hit in hits:
                frame_result = hit['_source'][frame]
                if 'validation_errors' in hit['_source'] and 'validation_errors' not in frame_result:
                    frame_result['validation_errors'] = hit['_source']['validation_errors']
                if 'aggregated_items' in hit['_source'] and 'aggregated_items' not in frame_result:
                    frame_result['aggregated_items'] = hit['_source']['aggregated_items']
                yield frame_result
            return

    def get_all_subsequent_results(self, extra_requests_needed_count, size_increment):
        """
        Calls `execute_search` in paginated manner iteratively until all results have been yielded.
        """
        from_ = 0
        while extra_requests_needed_count > 0:
            # print(str(extra_requests_needed_count) + " requests left to get all results.")
            from_ = from_ + size_increment
            subsequent_search_result = execute_search(es=self.es, query=self.query, index=self.es_index,
                                                      from_=from_, size=size_increment,
                                                      session_id=self.search_session_id)
            extra_requests_needed_count -= 1
            for hit in subsequent_search_result['hits'].get('hits', []):
                yield hit

    def execute_search_for_all_results(self, size_increment=100):
        """
        Returns a generator that iterates over _all_ results for search.
        Calls `execute_search` in paginated manner iteratively until all results have been yielded
        via `get_all_subsequent_results`.

        :param size_increment: number of results to get per page, default 100
        :return: all es_results that matched the given query
        """
        es_result = execute_search(es=self.es, query=self.query, index=self.es_index, from_=0, size=size_increment,
                                   session_id=self.search_session_id)

        total_results_expected = es_result['hits'].get('total', {}).get('value', 0)

        # Decrease by 1 (first es_result already happened)
        extra_requests_needed_count = math.ceil(total_results_expected / size_increment) - 1

        if extra_requests_needed_count > 0:
            # Returns a generator as value of es_result['hits']['hits']
            # Will be returned directly if self.return_generator is true
            # or converted to list if meant to be HTTP response.
            # Theoretical but unnecessary future: Consider allowing to return HTTP Stream of results w. return_generator=true (?)
            es_result['hits']['hits'] = itertools.chain(
                es_result['hits']['hits'],
                self.get_all_subsequent_results(extra_requests_needed_count, size_increment)
            )
        return es_result

    def execute_search(self):
        """ Executes the search, accounting for size if necessary """
        LuceneBuilder.verify_search_has_permissions(self.request, self.query)
        if self.size == 'all':
            es_results = self.execute_search_for_all_results()
        else:  # from_, size are integers
            es_results = execute_search(es=self.es, query=self.query, index=self.es_index,
                                        from_=self.from_, size=self.size,
                                        session_id=self.search_session_id)
        return es_results

    def format_results(self, es_results):
        """ Takes es_results from Elasticsearch and populates a response object based on
            on the given results.

        :param es_results: search results (from elasticsearch-dsl)
        """
        # Response formatting
        self.response['notification'] = 'Success'
        self.response['total'] = es_results['hits']['total']['value']
        self.response['facets'] = self.format_facets(es_results)
        self.response['aggregations'] = self.format_extra_aggregations(es_results)
        self.response['actions'] = self.get_collection_actions()
        columns = self.build_table_columns()
        if columns:
            self.response['columns'] = columns

        if self.size not in (None, 'all') and self.size < self.response['total']:
            params = [(k, v) for k, v in self.request.normalized_params.items() if k != 'limit']
            params.append(('limit', 'all'))
            # do not check presence of this field, as it triggers an ES len call! - Will 30 March 2022
            # if self.context:
            #     self.response['all'] = '%s?%s' % (self.request.resource_path(self.context), urlencode(params))

        # `graph` below is a generator.
        # `es_results['hits']['hits']` will contain a generator instead of list
        # if limit=all was requested. `self._format_results` will always return a generator
        # that iterates over es_results['hits']['hits'] regardless of its structure.
        graph = self._format_results(es_results['hits']['hits'])

        if self.return_generator:
            # Preserve `graph` as generator.
            self.response['@graph'] = graph
        else:
            # Convert it into list as we assume we're responding to a HTTP request by default.
            self.response['@graph'] = list(graph)

        # Save session ID for re-requests / subsequent pages.
        if self.search_session_id:  # Is 'None' if e.g. limit=all
            self.request.response.set_cookie('searchSessionID', self.search_session_id)

    def _sort_custom_facets(self):
        """ Applies custom sort to facets based on a dictionary provided on the type definition

            Specify a 2-tiered dictionary mapping field names to dictionaries of key -> weight
            mappings that allow us to sort generally like this:
                sorted(unsorted_terms, key=lambda d: field_terms_override_order.get(d['key'], default))
            ex:
            {
                {
                    facet_field_name: {
                        key1: weight,
                        key2: weight,
                        key3: weight
                        '_default': default_weight
                    }
                }
            }
            If you had field name and wanted to force a facet ordering, you
            could add this to the type definition:
                FACET_ORDER_OVERRIDE = {
                    'name': {
                        'Will': 1,
                        'Bob': 2,
                        'Alice': 3,
                        '_default': 4,
                    }
                }
            When faceting on the 'name' field, the ordering now will always be Will -> Bob -> Alice -> anything else
            regardless of the actual facet counts. Note that if no default is specified weight 101
            will be assigned (MAX_FACET_COUNTS + 1).
        """
        if 'facets' in self.response:
            for entry in self.response['facets']:
                field = entry.get('field')
                if field in self.facet_order_overrides:
                    field_terms_override_order = self.facet_order_overrides[field]
                    default = field_terms_override_order.get('_default', MAX_FACET_COUNTS + 1)
                    unsorted_terms = entry.get('terms', [])
                    entry['terms'] = sorted(unsorted_terms, key=lambda d: field_terms_override_order.get(d['key'],
                                                                                                         default))

    def get_response(self):
        """ Gets the response for this search, setting 404 status if necessary. """
        if not self.response:
            return {}  # XXX: rather than raise exception? -Will

        # add query if an admin asks for it
        if self.debug_is_active and is_admin_request(self.request):
            self.response['query'] = self.query

        # If we got no results, return 404 or []
        if not self.response['total']:
            # http://googlewebmastercentral.blogspot.com/2014/02/faceted-navigation-best-and-5-of-worst.html
            self.request.response.status_code = 404
            self.response['notification'] = 'No results found'
            self.response['@graph'] = []
            return self.response if not self.return_generator else []

        # if this is a subrequest/gen request, return '@graph' directly
        if self.request.__parent__ is not None or self.return_generator:
            if self.return_generator:
                # If self.return_generator, then self.response['@graph'] will
                # contain a generator rather than a list via `self.format_results`
                # TODO: Move that functionality into this method instead?
                return self.response['@graph']

        # apply custom facet filtering
        self._sort_custom_facets()

        # otherwise just hand off response
        return self.response

    def _build_query(self):
        """ Builds the query, setting self.query """
        self.initialize_search_response()
        self.build_search_query()

    def get_query(self):
        """ Grabs the query object, now a dictionary """
        return self.query

    def _search(self):
        """ Executes the end-to-end search.

        :returns: a search response (based on the __init__ parameters)
        """
        self._build_query()
        es_results = self.execute_search()
        self.format_results(es_results)
        return self.get_response()


@view_config(route_name='search', request_method='GET', permission='search')
@debug_log
def search(context, request, search_type=None, return_generator=False, forced_type='Search', custom_aggregations=None):
    """ Search view connects to ElasticSearch and returns the results """
    search_builder = SearchBuilder(context, request, search_type, return_generator, forced_type, custom_aggregations)
    return search_builder._search()


@view_config(context=AbstractCollection, permission='list', request_method='GET')
@debug_log
def collection_view(context, request):
    """
    Simply use search results for collections views (e.g./biosamples/)
    This is a redirect directly to the search page
    """
    return search(context, request, context.type_info.name, False, forced_type='Search')


def get_iterable_search_results(request, search_path='/search/', param_lists=None, inherit_user=True, **kwargs):
    '''
    Loops through search results, returns 100 (or search_results_chunk_row_size) results at a time. Pass it through itertools.chain.from_iterable to get one big iterable of results.
    Potential TODO: Move to search_utils or other file, and have this (or another version of this) handle compound filter_sets.

    :param request: Only needed to pass to do_subreq to make a subrequest with.
    :param search_path: Root path to call, defaults to /search/.
    :param param_lists: Dictionary of param:lists_of_vals which is converted to URL query.
    :param search_results_chunk_row_size: Amount of results to get per chunk. Default should be fine.
    '''
    param_lists = deepcopy(param_lists)
    param_lists['limit'] = ['all']
    param_lists['from'] = [0]
    param_lists['sort'] = param_lists.get('sort', 'uuid')
    subreq = make_search_subreq(request, '{}?{}'.format(search_path, urlencode(param_lists, True)), inherit_user=inherit_user)
    if not inherit_user:
        # Perform request as if an admin.
        subreq.remote_user = "UPGRADE"
        if 'HTTP_COOKIE' in subreq.environ:
            del subreq.environ['HTTP_COOKIE']
    return iter_search_results(None, subreq, **kwargs)


# Update? used in ./batch_download.py
def iter_search_results(context, request, **kwargs):
    return search(context, request, return_generator=True, **kwargs)


_ASSEMBLY_MAPPER = {
    'GRCh38-minimal': 'hg38',
    'GRCh38': 'hg38',
    'GRCh37': 'hg19',
    'GRCm38': 'mm10',
    'GRCm37': 'mm9',
    'BDGP6': 'dm4',
    'BDGP5': 'dm3',
    'WBcel235': 'WBcel235'
}

hgConnect = 'http://genome.ucsc.edu/cgi-bin/hgTracks?hubClear='



"""
Exception definitions for ingestion
"""

from pyramid.httpexceptions import HTTPBadRequest, HTTPServerError


class SubmissionFailure(HTTPServerError):
    pass


class UndefinedIngestionProcessorType(Exception):

    def __init__(self, processor_type):
        self.ingestion_type_name = processor_type
        super().__init__("No ingestion processor type %r is defined." % processor_type)


class MissingParameter(HTTPBadRequest):

    def __init__(self, parameter_name):
        self.parameter_name = parameter_name
        super().__init__(detail="Missing parameter: %s" % parameter_name)


class BadParameter(HTTPBadRequest):

    def __init__(self, parameter_name, parameter_value, extra_detail=None):
        self.parameter_name = parameter_name
        self.parameter_value = parameter_value
        self.extra_detail = extra_detail
        suffix = " " + extra_detail if extra_detail else ""
        super().__init__(detail="The value of the %s parameter, %r, is invalid.%s"
                                % (parameter_name, parameter_value, suffix))


class UnspecifiedFormParameter(HTTPBadRequest):

    def __init__(self, parameter_name):
        self.parameter_name = parameter_name
        super().__init__(detail="A form parameter was not filled out: %s" % parameter_name)



import sys
import argparse
import logging
import structlog
import transaction
from pyramid.paster import get_app

from dcicutils.env_utils import is_stg_or_prd_env
from .. import STORAGE
from ..elasticsearch.indexer_utils import get_uuids_for_types
from .. import configure_dbsession


logger = structlog.getLogger(__name__)
EPILOG = __doc__


def purge_item_type_from_storage(app, item_types, prod=False):
    """
    Purges all items with the given item_types from our storage. This function could partially purge
    an item type if an error is encountered. Note that this will work no matter what resources are backing
    'PickStorage'.

    IMPORTANT: If an error occurs the DB transaction is rolled back, but the ES deletions will persist.
               Re-running 'create_mapping' on this item type will reindex the items.

    :param app: app to access settings from, either a testapp (testapp.app.registry) or regular app(.registry)
    :param item_types: list of types to purge from DB
    :param prod: bool whether to allow run on prod, default False
    :return: True in success, False otherwise
    """
    if not hasattr(app, 'registry'):
        app = app.app
        if not hasattr(app, 'registry'):
            raise RuntimeError('Passed app to purge_item_type_from_db does not contain a registry.')

    if 'env.name' in app.registry.settings:
        env = app.registry.settings['env.name']
        if is_stg_or_prd_env(env) and not prod:
            logger.error('Tried to run purge_item_type_from_storage on prod without specifying'
                         'the prod options - exiting.')
            return False

    # purge uuids directly from PickStorage, ignoring status=deleted checks
    configure_dbsession(app)
    # The use of set(...) is to de-duplicate. -kmp 28-Jan-2021
    uuids_to_purge = set(get_uuids_for_types(app.registry, item_types))
    pstorage = app.registry[STORAGE]
    for uuid in uuids_to_purge:
        try:
            pstorage.purge_uuid(uuid)
            transaction.commit()
        except Exception as e:  # XXX: handle recoverable exceptions?
            logger.error('Encountered exception purging an item type (uuid: %s) from the DB: %s'
                         % (uuid, e))
            transaction.abort()
            return False

    return True


def main():
    """ Entry point for this command """
    logging.basicConfig()

    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description='Clear an item type out of metadata storage',
        epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument('config_uri', help='path to configfile')
    parser.add_argument('item_type', help='item type to run on')
    parser.add_argument('--app-name', help='Pyramid app name in configfile')
    parser.add_argument('--prod', help='Whether or not to proceed if we are on a production server',
                        action='store_true', default=False)
    args = parser.parse_args()

    app = get_app(args.config_uri, args.app_name)
    sys.exit(purge_item_type_from_storage(app, [args.item_type], prod=args.prod))


if __name__ == '__main__':
    main()



from dcicutils.project_utils import C4ProjectRegistry, C4Project
from .project.access_key import SnovaultProjectAccessKey
from .project.authentication import SnovaultProjectAuthentication
from .project.authorization import SnovaultProjectAuthorization
from .project.ingestion import SnovaultProjectIngestion
from .project.loadxl import SnovaultProjectLoadxl
from .project.renderers import SnovaultProjectRenderers


@C4ProjectRegistry.register("dcicsnovault")
class SnovaultProject(SnovaultProjectAccessKey,
                      SnovaultProjectAuthentication,
                      SnovaultProjectAuthorization,
                      SnovaultProjectIngestion,
                      SnovaultProjectLoadxl,
                      SnovaultProjectRenderers,
                      C4Project):
    NAMES = {"NAME": "snovault", "PYPI_NAME": "dcicsnovault"}
    ACCESSION_PREFIX = "SNO"



import pytest
from snovault.ingestion.ingestion_processor_decorator import ingestion_processor
from snovault.types.ingestion import SubmissionFolio


def test_error_ingestion_processor():
    with pytest.raises(Exception):
        @ingestion_processor('some_unknown_ingestion_type')
        def some_processor_for_unknown_ingestion_type(submission: SubmissionFolio):
            pass



import pytest  # noQA
from snovault.schema_validation import NO_DEFAULT
from snovault.schema_utils import SchemaValidator
from jsonschema.exceptions import ValidationError


validator_class = SchemaValidator


fake_schema = {
    '$id': 'abc',
    'title': 'Fake',
    'description': 'Schema',
    '$schema': 'https://json-schema.org/draft/2020-12/schema',
    'type': 'object',
    'required': ['award', 'lab'],
    'identifyingProperties': ['uuid'],
    'additionalProperties': False,
    'properties': {
        'uuid': {
            'type': 'string'
        },
        'award': {
            'type': 'string'
        },
        'lab': {
            'type': 'string'
        }
    }
}


def make_default(instance, subschema):
    if instance.get('skip'):
        return NO_DEFAULT
    assert subschema == {'serverDefault': 'test'}
    return 'bar'


def test_validator_serializes_default_properties():
    schema = {
        'properties': {
            'foo': {
                'default': 'bar'
            }
        }
    }
    result, errors = validator_class(
        schema
    ).serialize(
        {}
    )
    assert result == {'foo': 'bar'}
    assert errors == []


def test_validator_serializes_default_properties_in_items():
    schema = {
        'items': {
            'properties': {
                'foo': {
                    'default': 'bar'
                }
            }
        }
    }
    result, errors = validator_class(
        schema,
    ).serialize(
        [
            {}
        ]
    )
    assert result == [{'foo': 'bar'}]
    assert errors == []


def test_validator_serializes_server_default_properties():
    schema = {
        'properties': {
            'foo': {
                'serverDefault': 'test'
            }
        }
    }
    result, errors = validator_class(
        schema,
    ).add_server_defaults(
        {
            'test': make_default
        }
    ).serialize(
        {}
    )
    assert result == {'foo': 'bar'}
    assert errors == []


def test_validator_ignores_server_default_returning_no_default():
    schema = {
        'properties': {
            'foo': {
                'serverDefault': 'test'
            },
            'skip': {}
        }
    }
    result, errors = validator_class(
        schema,
    ).add_server_defaults(
        {
            'test': make_default
        }
    ).serialize(
        {
            'skip': True
        }
    )
    assert result == {'skip': True}
    assert errors == []


def test_validator_serializes_server_default_properties_in_items():
    schema = {
        'items': {
            'properties': {
                'foo': {
                    'serverDefault': 'test'
                }
            }
        }
    }
    result, errors = SchemaValidator(
        schema
    ).add_server_defaults(
        {
            'test': make_default
        }
    ).serialize(
        [{}]
    )
    assert result == [{'foo': 'bar'}]
    assert errors == []


def test_validator_serializes_properties_in_order_of_input():
    schema = {
        'properties': {
            'foo': {},
            'bar': {},
        }
    }
    validator = validator_class(
        schema
    )
    value = {
        'bar': 1,
        'foo': 2
    }
    result, errors = validator.serialize(value)
    assert list(result) == ['bar', 'foo']
    assert errors == []


def test_validator_returns_error():
    schema = {
        'type': 'object',
        'properties': {
            'name': {
                'type': 'string'
            }
        },
        'required': ['name']
    }
    result, errors = validator_class(
        schema,
    ).serialize(
        {
            'name': 'abc'
        }
    )
    assert result == {'name': 'abc'}
    assert not errors
    result, errors = validator_class(
        schema,
    ).serialize(
        {
            'name': 1
        }
    )
    assert result == {
        'name': 1
    }
    assert isinstance(errors[0], ValidationError)
    assert errors[0].message == "1 is not of type 'string'"
    result, errors = validator_class(
        schema,
    ).serialize(
        {}
    )
    assert result == {}
    assert isinstance(errors[0], ValidationError)
    assert errors[0].message == "'name' is a required property"


def test_validator_check_schema():
    validator_class.check_schema(fake_schema)


def test_validator_extend_with_default():
    from copy import deepcopy
    from snovault.schema_validation import SerializingSchemaValidator
    original_instance = {
        'x': 'y'
    }
    mutated_instance = deepcopy(original_instance)
    assert original_instance == mutated_instance
    schema = {'properties': {'foo': {'default': 'bar'}}}
    SerializingSchemaValidator(schema).validate(mutated_instance)
    assert original_instance == {'x': 'y'}
    assert mutated_instance == {'x': 'y', 'foo': 'bar'}


def test_validator_extend_with_default_and_serialize():
    instance = {
        'x': 'y'
    }
    from snovault.schema_validation import SerializingSchemaValidator
    schema = {'properties': {'foo': {'default': 'bar'}}}
    result, errors = SerializingSchemaValidator(schema).serialize(instance)
    assert instance == {'x': 'y'}
    assert result == {'x': 'y', 'foo': 'bar'}
    assert errors == []
    schema = {
        'properties': {
            'foo': {
                'default': 'bar'
            },
            'name': {
                'type': 'string'
            }
        },
        'required': ['name']
    }
    result, errors = SerializingSchemaValidator(schema).serialize(
        {
            'foo': 'thing',
        }
    )
    assert result == {'foo': 'thing'}
    assert errors[0].message == "'name' is a required property"


def test_validator_extend_with_server_default_and_serialize():
    instance = {
        'x': 'y'
    }
    from snovault.schema_validation import SerializingSchemaValidator
    schema = {'properties': {'foo': {'serverDefault': 'test'}}}
    result, errors = SerializingSchemaValidator(
        schema,
    ).add_server_defaults(
        {
            'test': make_default
        }
    ).serialize(instance)
    assert instance == {'x': 'y'}
    assert result == {'x': 'y', 'foo': 'bar'}
    assert errors == []
    schema = {
        'properties': {
            'foo': {
                'serverDefault': 'test'
            },
            'name': {
                'type': 'string'
            }
        },
        'required': ['name']
    }
    result, errors = SerializingSchemaValidator(
        schema
    ).add_server_defaults(
        {
            'test': make_default
        }
    ).serialize(
        {
            'foo': 'thing',
        }
    )
    assert result == {'foo': 'thing'}
    assert errors[0].message == "'name' is a required property"
    result, errors = SerializingSchemaValidator(
        schema,
    ).add_server_defaults(
        {
            'test': make_default
        }
    ).serialize(
        {
            'name': 'other thing',
        }
    )
    assert result == {
        'foo': 'bar',
        'name': 'other thing',
    }
    assert not errors
